"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Sequence, Dict, Tuple, Callable

import torch

from fuse.models.backbones.backbone_inception_resnet_v2 import FuseBackboneInceptionResnetV2
from fuse.models.heads.head_global_pooling_classifier import FuseHeadGlobalPoolingClassifier
from fuse.utils.utils_hierarchical_dict import FuseUtilsHierarchicalDict


class FuseModelWrapper(torch.nn.Module):
    """
    Fuse model wrapper for wrapping torch modules and passing through Fuse
    """

    def __init__(self,
                 model: torch.nn.Module = None,
                 model_inputs: Sequence[str] = ('data.input.input_0.tensor',),
                 pre_forward_processing_function: Callable = None,
                 post_forward_processing_function: Callable = None,
                 model_outputs: Sequence[str] = ('output.output_0',)
                 ) -> None:
        """
        Default Fuse model wrapper - extracts batch_dict data from model_inputs keys and calls forward of model.
        Then, puts forward's output into batch_dict model_output keys.
        For pre processing of the batch dict data use pre_forward_processing_function.
        For post processing of the forward's output data use post_forward_processing_function.

        :param model: The model to wrap
        :param model_inputs: sequence of keys in batch dict to transfer into model.forward function
        :param pre_forward_processing_function: utility function to process input data before forward is called (after it's extracted from batch_dict)
        :param post_forward_processing_function: utility function to process forward's output data before it is saved into batch_dict
        :param model_outputs: keys in batch dict to save the outputs of model.forward()

        """
        super().__init__()
        self.model_inputs = model_inputs
        self.model = model
        self.add_module('wrapped_model', self.model)
        self.pre_forward_processing_function = pre_forward_processing_function
        self.post_forward_processing_function = post_forward_processing_function
        self.model_outputs = model_outputs

    def forward(self,
                batch_dict: Dict) -> Dict:
        # convert input to the model's expected input
        model_input = [FuseUtilsHierarchicalDict.get(batch_dict, conv_input) for conv_input in self.model_inputs]

        # convert input to model expected input
        if self.pre_forward_processing_function is not None:
            model_input = self.pre_forward_processing_function(model_input)

        # run the model
        model_output = self.model.forward(*model_input)

        # convert output of model to Fuse expected output
        if self.post_forward_processing_function is not None:
            model_output = self.post_forward_processing_function(model_output)

        if len(self.model_outputs) == 1:
            FuseUtilsHierarchicalDict.set(batch_dict, 'model.' + self.model_outputs[0], model_output)
        else:
            for i, output_name in enumerate(self.model_outputs):
                FuseUtilsHierarchicalDict.set(batch_dict, 'model.' + output_name, model_output[i])

        return batch_dict['model']


if __name__ == '__main__':
    from torchvision.models.googlenet import BasicConv2d
    import torchvision.models as models
    import torch
    import os

    os.environ['CUDA_VISIBLE_DEVICES'] = "0, 1"

    DEVICE = 'cpu'  # 'cuda'
    DATAPARALLEL = False  # True

    googlemet_model = models.googlenet.GoogLeNet()
    googlemet_model.conv1 = BasicConv2d(1, 64, kernel_size=7, stride=2, padding=3)

    def convert_googlenet_outputs(output):
        return output.logits

    model = FuseModelWrapper(
        model_inputs=['data.input.input_0.tensor'],
        model=googlemet_model,
        post_forward_processing_function=convert_googlenet_outputs,
        model_outputs=['google.output.logits']
    )

    model = model.to(DEVICE)
    if DATAPARALLEL:
        model = torch.nn.DataParallel(model)

    dummy_data = {'data':
                      {'input':
                           {'input_0': {'tensor': torch.zeros([17, 1, 200, 100]).to(DEVICE), 'metadata': None}},
                       'gt':
                           {'gt_0': torch.zeros([0]).to(DEVICE)}
                       }
                  }

    res = {}
    res['model'] = model.forward(dummy_data)
    print('Forward pass shape: ', end='')
    print('logits', str(res['model']['google']['output']['logits'].shape))




"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

"""
Dummy cache implementation, doing nothing
"""
from multiprocessing import Manager
from typing import Hashable, Any, List

from fuse.data.cache.cache_base import FuseCacheBase


class FuseCacheNull(FuseCacheBase):
    def __init__(self):
        super().__init__()

    def __contains__(self, key: Hashable) -> bool:
        """
        See base class
        """
        return False

    def __getitem__(self, key: Hashable) -> Any:
        """
        See base class
        """
        return None

    def __delitem__(self, key: Hashable) -> None:
        """
        See base clas
        """
        pass

    def __setitem__(self, key: Hashable, value: Any) -> None:
        """
        See base class
        """
        pass

    def save(self) -> None:
        """
        See base class
        """
        pass

    def exist(self) -> bool:
        """
        See base class
        """
        return True

    def reset(self) -> None:
        """
        See base class
        """
        pass

    def get_all_keys(self, include_none: bool = False) -> List[Hashable]:
        """
        See base class
        """
        return []

    def start_caching(self, manager: Manager):
        """
        See base class
        """
        pass



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import shutil
import tempfile
import unittest
import os

from fuse.utils.utils_gpu import FuseUtilsGPU
from fuse_examples.classification.mnist.runner import TRAIN_COMMON_PARAMS, run_train, run_infer, run_analyze, INFER_COMMON_PARAMS, \
    ANALYZE_COMMON_PARAMS


class ClassificationMnistTestCase(unittest.TestCase):

    def setUp(self):

        self.root = tempfile.mkdtemp()

        self.paths = {
            'model_dir': os.path.join(self.root, 'mnist/model_dir'),
            'force_reset_model_dir': True,  # If True will reset model dir automatically - otherwise will prompt 'are you sure' message.
            'cache_dir': os.path.join(self.root, 'mnist/cache_dir'),
            'inference_dir': os.path.join(self.root, 'mnist/infer_dir'),
            'analyze_dir': os.path.join(self.root, 'mnist/analyze_dir')}


        self.train_common_params = TRAIN_COMMON_PARAMS

        self.infer_common_params = INFER_COMMON_PARAMS

        self.analyze_common_params = ANALYZE_COMMON_PARAMS

    def test_template(self):
        num_gpus_allocated = FuseUtilsGPU.choose_and_enable_multiple_gpus(1, use_cpu_if_fail=True)
        if num_gpus_allocated == 0:
            self.train_common_params['manager.train_params']['device'] = 'cpu'
        run_train(self.paths, self.train_common_params)
        run_infer(self.paths, self.infer_common_params)
        results = run_analyze(self.paths, self.analyze_common_params)

        threshold = 0.98
        self.assertGreaterEqual(results['auc']['macro_avg'], threshold)

    def tearDown(self):
        # Delete temporary directories
        shutil.rmtree(self.root)


if __name__ == '__main__':
    unittest.main()


"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

"""
Fuse Dataset Base
"""
import pickle
from abc import abstractmethod
from enum import Enum
from typing import Any, List, Optional

from torch.utils.data.dataset import Dataset


class FuseDatasetBase(Dataset):
    """
    Abstract base class for Fuse dataset.
    All subclasses should overwrite the following abstract methods inherited from  torch.utils.data.Dataset
    `__getitem__`, supporting fetching a data sample for a given key.
    `__len__`, which is expected to return the size of the dataset
    And the ones listed below
    """

    class SaveMode(Enum):
        # store just the info required for inference
        INFERENCE = 1,
        # store all the info
        TRAINING = 2

    def __init__(self):
        super().__init__()

    @abstractmethod
    def create(self, **kwargs) -> None:
        """
        Used to enable the instance
        Typically will load caching, etc
        :param kwargs: different parameters per subclass
        :return: None
        """
        raise NotImplementedError

    @abstractmethod
    def get(self, index: Optional[int], key: Optional[str], use_cache: bool = False) -> Any:
        """
        Get input, ground truth or metadata of a sample.

        :param index: the index of the item or None for all
        :param key: string representing the exact information required, use None for all.
        :param use_cache: if true, will try to reload the sample from caching mechanism in case exist.
        :return: the required info of a single sample of a list of samples
        """
        raise NotImplementedError

    @abstractmethod
    def collate_fn(self, samples: List[Any]) -> Any:
        """
        collate list of samples into batch
        :param samples: list of samples
        :return: batch
        """
        raise NotImplementedError

    # misc
    @abstractmethod
    def summary(self, statistic_keys: Optional[List[str]] = None) -> str:
        """
        String summary of the object
        :param statistic_keys: Optional. list of keys to output statistics about.
        """
        raise NotImplementedError

    # save and load datasets
    @abstractmethod
    def get_instance_to_save(self, mode: SaveMode) -> 'FuseDatasetBase':
        """
        Create lite instance version of dataset with just the info required to recreate it
        :param mode: see SaveMode for available modes
        :return: the instance to save
        """
        raise NotImplementedError

    @staticmethod
    def save(dataset: 'FuseDatasetBase', mode: SaveMode, filename: str) -> None:
        """
        Static method save dataset to the disc (see SaveMode for available modes)
        :param dataset: the dataset to save
        :param mode: required mode to save
        :param filename: file name to use
        :return: None
        """
        # get instance version to save
        dataset_to_save = dataset.get_instance_to_save(mode)

        # save this instance
        with open(filename, 'wb') as pickle_file:
            pickle.dump(dataset_to_save, pickle_file)

    @staticmethod
    def load(filename: str, **kwargs) -> 'FuseDatasetBase':
        """
        load dataset
        :param filename: path to saved dataset
        :param kwargs: arguments of create() function
        :return: the dataset object
        """
        # load saved instance
        with open(filename, 'rb') as pickle_file:
            dataset = pickle.load(pickle_file)

        # recreate dataset
        dataset.create(**kwargs)

        return dataset



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import secrets
from abc import ABC, abstractmethod
from typing import Any, Sequence, List, Tuple

import torch
from torch import Tensor

secretsGenerator = secrets.SystemRandom()


class FuseUtilsParamSamplerBase(ABC):
    """
    Base class for param sampler
    """

    @abstractmethod
    def sample(self) -> Any:
        """
        :return: the sampled value
        """
        raise NotImplementedError


class FuseUtilsParamSamplerUniform(FuseUtilsParamSamplerBase):
    def __init__(self, min: float, max: float):
        """
        Uniform distribution between min and max
        """
        super().__init__()

        # store input parameters
        self.min = min
        self.max = max

    def sample(self) -> float:
        """
        :return: random float in range [min-max]
        """
        return secretsGenerator.uniform(self.min, self.max)

    def __str__(self):
        return f'RandUniform [{self.min} - {self.max}] '


class FuseUtilsParamSamplerRandInt(FuseUtilsParamSamplerBase):
    def __init__(self, min: int, max: int):
        """
        Uniform integer distribution between min and max
        """
        super().__init__()

        # store input parameters
        self.min = min
        self.max = max

    def sample(self) -> int:
        """
        :return: random int in range [min-max]
        """
        return secretsGenerator.randint(self.min, self.max)

    def __str__(self):
        return f'RandInt [{self.min} - {self.max}] '


class FuseUtilsParamSamplerRandBool(FuseUtilsParamSamplerBase):
    def __init__(self, probability: float):
        """
        Random boolean according to the given probability
        """
        super().__init__()

        # store input parameters
        self.probability = probability

    def sample(self) -> bool:
        """
        :return: random boolean according to probability
        """
        return secretsGenerator.uniform(0, 1) <= self.probability

    def __str__(self):
        return f'RandBool p={self.probability}] '


class FuseUtilsParamSamplerChoice(FuseUtilsParamSamplerBase):

    def __init__(self, seq: Sequence, probabilities: List[float], k: int = 0):
        """
        Random choice out of a sequence
        Return a k sized list of population elements chosen with replacement
        If k=0 return a single element instead of a list
        """
        super().__init__()

        # store input parameters
        self.seq = seq
        self.probabilities = probabilities
        self.k = k

    def sample(self) -> Any:
        """
        :return: random element according to probabilities
        """
        if self.k == 0:
            return secretsGenerator.choices(self.seq, weights=self.probabilities)[0]
        else:
            return secretsGenerator.choices(self.seq, weights=self.probabilities, k=self.k)

    def __str__(self):
        return f'RandChoice seq={self.seq}, w={self.probabilities}] '


def sample_all(data: Any) -> Any:
    """
    Generate a copy of the data structure, replacing each FuseUtilsParamSamplerBase with a random sample.

    :param data:data_structure: recursivelly looking for FuseUtilsParamSamplerBase in a dictionary and a sequence
    :return: See above
    """
    # if a dictionary return a copy of the dictionary try to sample each value recursively
    if isinstance(data, dict):
        data_dict: dict = data.copy()
        for key in data_dict:
            data_dict[key] = sample_all(data_dict[key])
        return data_dict

    # if a list  return a copy of the list try to sample each element recursively
    if isinstance(data, list):
        data_lst: list = data[:]
        for ii in range(len(data_lst)):
            data_lst[ii] = sample_all(data_lst[ii])
        return data_lst

    # if a tuple  return a copy of the tuple try to sample each element recursively
    if isinstance(data, Tuple):
        data_tuple = tuple((sample_all(data[ii]) for ii in range(len(data))))
        return data_tuple

    # if FuseUtilsParamSamplerBase, sample a number
    if isinstance(data, FuseUtilsParamSamplerBase):
        data_sampler: FuseUtilsParamSamplerBase = data
        return data_sampler.sample()

    # otherwise return the original data
    return data


class FuseUtilsParamSamplerGaussianPatch(FuseUtilsParamSamplerBase):
    def __init__(self, shape: Tuple[int, ...], mean: float, std: float):
        """
        Gaussian noise
        """
        super().__init__()

        # store input parameters
        self.shape = shape
        self.mean = mean
        self.std = std

    def sample(self) -> Tensor:
        """
        :return: random gaussian noise
        """
        return self.std * torch.randn(self.shape) + self.mean



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import logging
from datetime import datetime
from typing import Dict

import torch.nn as nn

from fuse.managers.callbacks.callback_base import FuseCallback
from fuse.utils.utils_hierarchical_dict import FuseUtilsHierarchicalDict
from fuse.utils.utils_misc import FuseUtilsMisc


class FuseCallbackDebug(FuseCallback):
    """
    Callback used to log the information about each stage: begin/end time and fuse main structure: batch_dict, virtual_batch_results and
    epoch_results
    """

    def __init__(self):
        super().__init__()
        self.time_step_begin = None
        self.time_epoch_begin = None
        self.time_virtual_batch_begin = None
        self.time_batch_begin = None

        self.first_step = True
        self.first_epoch = {}
        self.first_virtual_batch = {}
        self.first_batch = {}

    def on_step_begin(self, step: int) -> None:
        # log info on first step only
        if self.first_step:
            self.time_step_begin = datetime.now()
            current_time = self.time_step_begin.strftime("%H:%M:%S")
            logging.getLogger('Fuse').info(f'\nStep {step} - {current_time} - BEGIN', {'color': 'green', 'attrs': ['bold', 'underline']})

    def on_step_end(self, step: int, train_results: Dict = None, validation_results: Dict = None, learning_rate: float = None) -> None:
        # log info on first step only: learning rate, train results and validation results
        if self.first_step:
            lgr = logging.getLogger('Fuse')

            time_step_end = datetime.now()
            current_time = time_step_end.strftime("%H:%M:%S")
            total_time = time_step_end - self.time_step_begin
            lgr.info(f'Step {step} - {current_time} (total time {total_time}) - END', {'color': 'green', 'attrs': ['bold', 'underline']})

            lgr.info(f'Step Results:', {'color': 'green', 'attrs': 'bold'})
            lgr.info(f'Learning Rate: {learning_rate}', {'color': 'green'})
            lgr.info(f'Train Results:', {'color': 'green', 'attrs': 'bold'})
            lgr.info(f'{FuseUtilsHierarchicalDict.to_string(train_results)}', {'color': 'green'})
            lgr.info(f'Validation Results:', {'color': 'green', 'attrs': 'bold'})
            lgr.info(f'{FuseUtilsHierarchicalDict.to_string(validation_results)}', {'color': 'green'})

            self.time_step_begin = None
            self.first_step = False

    def on_epoch_begin(self, mode: str, epoch: int) -> None:
        # log info on first train/validation epoch only
        if self.first_epoch.get(mode, True):
            self.time_epoch_begin = datetime.now()
            current_time = self.time_epoch_begin.strftime("%H:%M:%S")
            logging.getLogger('Fuse').info(f'\n{mode.capitalize()} epoch {epoch} - {current_time} - BEGIN', {'color': 'green', 'attrs': ['bold', 'underline']})

    def on_epoch_end(self, mode: str, epoch: int, epoch_results: Dict = None) -> None:
        # log info on first train/validation epoch only: epoch results
        if self.first_epoch.get(mode, True):
            lgr = logging.getLogger('Fuse')

            time_epoch_end = datetime.now()
            current_time = time_epoch_end.strftime("%H:%M:%S")
            total_time = time_epoch_end - self.time_epoch_begin
            lgr.info(f'{mode.capitalize()}  epoch {epoch} - {current_time} (total time {total_time}) - END', {'color': 'green', 'attrs': ['bold', 'underline']})

            lgr.info(f'Epoch Results:', {'color': 'green', 'attrs': 'bold'})
            lgr.info(f'{FuseUtilsHierarchicalDict.to_string(epoch_results)}', {'color': 'green'})

            self.time_epoch_begin = None
            self.first_epoch[mode] = False

    def on_virtual_batch_begin(self, mode: str, virtual_batch: int) -> None:
        # log info on first train/validation virtual batch only
        if self.first_virtual_batch.get(mode, True):
            self.time_virtual_batch_begin = datetime.now()
            current_time = self.time_virtual_batch_begin.strftime("%H:%M:%S")
            logging.getLogger('Fuse').info(f'\n{mode.capitalize()} virtual batch {virtual_batch} - {current_time} - BEGIN', {'color': 'green', 'attrs': ['bold', 'underline']})

    def on_virtual_batch_end(self, mode: str, virtual_batch: int, virtual_batch_results: Dict = None) -> None:
        # log info on first train/validation virtual batch only: virtual batch results
        if self.first_virtual_batch.get(mode, True):
            lgr = logging.getLogger('Fuse')

            time_virtual_batch_end = datetime.now()
            current_time = time_virtual_batch_end.strftime("%H:%M:%S")
            total_time = time_virtual_batch_end - self.time_virtual_batch_begin
            lgr.info(f'{mode.capitalize()} virtual batch {virtual_batch} - {current_time} (total time {total_time}) - END', {'color': 'green', 'attrs': ['bold', 'underline']})

            lgr.info(f'Virtual Batch Results:', {'color': 'green', 'attrs': 'bold'})
            lgr.info(f'{FuseUtilsHierarchicalDict.to_string(virtual_batch_results)}', {'color': 'green'})

            self.time_virtual_batch_begin = None
            self.first_virtual_batch[mode] = False

    def on_batch_begin(self, mode: str, batch: int) -> None:
        # log info on first train/validation batch only
        if self.first_batch.get(mode, True):
            self.time_batch_begin = datetime.now()
            current_time = self.time_batch_begin.strftime("%H:%M:%S")
            logging.getLogger('Fuse').info(f'\n{mode.capitalize()} batch {batch} - {current_time} - BEGIN', {'color': 'green', 'attrs': ['bold', 'underline']})

    def on_data_fetch_end(self, mode: str, batch: int, batch_dict: Dict = None) -> None:
        # log info on first train/validation batch only: batch_dict
        if self.first_batch.get(mode, True):
            lgr = logging.getLogger('Fuse')

            time_batch_end = datetime.now()
            current_time = time_batch_end.strftime("%H:%M:%S")
            total_time = time_batch_end - self.time_batch_begin
            lgr.info(f'{mode.capitalize()} data fetch {batch} - {current_time} (total time {total_time}) - END',
                     {'color': 'green', 'attrs': ['bold', 'underline']})

            lgr.info(f'Batch Dict:', {'color': 'green', 'attrs': 'bold'})
            lgr.info(f'{FuseUtilsMisc.batch_dict_to_string(batch_dict)}', {'color': 'green'})

    def on_batch_end(self, mode: str, batch: int, batch_dict: Dict = None) -> None:
        # log info on first train/validation batch only: batch_dict
        if self.first_batch.get(mode, True):
            lgr = logging.getLogger('Fuse')

            time_batch_end = datetime.now()
            current_time = time_batch_end.strftime("%H:%M:%S")
            total_time = time_batch_end - self.time_batch_begin
            lgr.info(f'{mode.capitalize()} batch {batch} - {current_time} (total time {total_time}) - END', {'color': 'green', 'attrs': ['bold', 'underline']})

            lgr.info(f'Batch Dict:', {'color': 'green', 'attrs': 'bold'})
            lgr.info(f'{FuseUtilsMisc.batch_dict_to_string(batch_dict)}', {'color': 'green'})

            self.time_batch_begin = None
            self.first_batch[mode] = False

    def on_train_begin(self, net: nn.Module) -> None:
        lgr = logging.getLogger('Fuse')
        lgr.info(f'\nTrain - BEGIN', {'color': 'green', 'attrs': ['bold', 'underline']})

    def on_train_end(self) -> None:
        lgr = logging.getLogger('Fuse')
        lgr.info(f'\nTrain - DONE', {'color': 'green', 'attrs': ['bold', 'underline']})



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import logging
import time
from typing import Dict

import torch.nn as nn

from fuse.managers.callbacks.callback_base import FuseCallback
from fuse.managers.manager_state import FuseManagerState
from fuse.utils import utils_misc as misc


class FuseTimeStatisticsCallback(FuseCallback):
    """
        Counts time of procedures.
    """

    def __init__(self, num_epochs: int, load_expected_part=0.1) -> None:
        """

        :param num_epochs: total number of epochs (to compute remaining time)
        :param load_expected_part: expected fraction of the loading from the total handle_epoch time.
        """
        super().__init__()
        self.num_epochs = num_epochs
        self.load_expected_part = load_expected_part

        self.step_begin_time = None
        self.epoch_begin_time = None
        self.batch_begin_time = None
        self.virtual_batch_begin_time = None
        self.load_batch_aggregated_time = 0
        self.train_begin_time = None

        pass

    def on_step_begin(self, step: int) -> None:
        self.step_begin_time = time.time()
        pass

    def on_step_end(self, step: int, train_results: Dict = None, validation_results: Dict = None, learning_rate: float = None) -> None:
        """
        Counts the time of a step (train + validation + update scheduler), and computes the remaining time for the entire run.

        :param step: step number (out of self.num_epochs)
        :param train_results: ignored
        :param validation_results: ignored
        :param learning_rate: ignored
        """
        end_time = time.time()
        timedelta_seconds = end_time - self.step_begin_time
        running_time_left_seconds = timedelta_seconds * (self.num_epochs - step)
        running_time_left_str = misc.time_display(running_time_left_seconds)
        epoch_time_str = misc.time_display(timedelta_seconds)
        logging.getLogger('Fuse').info(f'Estimated time left: {running_time_left_str} (last epoch time: {epoch_time_str})', {'color': 'cyan'})
        pass

    def on_epoch_begin(self, mode: str, epoch: int) -> None:
        """
        reset epoch_begin_time and the load_batch_aggregated time
        :param mode: ignored
        :param epoch: ignored
        """
        self.epoch_begin_time = time.time()
        self.load_batch_aggregated_time = 0
        pass

    def on_epoch_end(self, mode: str, epoch: int, epoch_results: Dict = None) -> None:
        """
        Writes the time for epoch, and checks whether the loading of data took more than load_expected_part.
        If the loading fraction is greater than expected, notifies logger.

        :param mode: mode
        :param epoch: epoch number
        :param epoch_results: ignored
        """
        epoch_end = time.time()

        lgr = logging.getLogger('Fuse')
        lgr.debug(f"Time for {mode} epoch {epoch}: {misc.get_time_delta(self.epoch_begin_time)}")

        # debug info about the batch iterator loading time
        load_time_str = misc.time_display(self.load_batch_aggregated_time)
        lgr.debug(f'Mode: {mode}, epoch {epoch}: Total time for loading the data is {load_time_str}')

        # check if the loading part of the data exceeded load_expected_part of epoch time
        epoch_time = epoch_end - self.epoch_begin_time
        max_time_for_load = epoch_time * self.load_expected_part
        if self.load_batch_aggregated_time > max_time_for_load:
            lgr.info(f'Mode: {mode}, epoch {epoch}: '
                     f'Total time for loading data ({load_time_str}) is greater than expected (maximum {misc.time_display(max_time_for_load)})',
                     {'color': 'blue'})
        pass

    def on_virtual_batch_begin(self, mode: str, virtual_batch: int) -> None:
        self.virtual_batch_begin_time = time.time()
        pass

    def on_virtual_batch_end(self, mode: str, virtual_batch: int, virtual_batch_results: Dict = None) -> None:
        logging.getLogger('Fuse').debug(f"Time for {mode} virtual batch {virtual_batch}: {misc.get_time_delta(self.virtual_batch_begin_time)}")
        pass

    def on_batch_begin(self, mode: str, batch: int) -> None:
        self.batch_begin_time = time.time()
        pass

    def on_data_fetch_end(self, mode: str, batch: int, batch_dict: Dict = None) -> None:
        # add the delta time to load to the aggregated sum
        self.load_batch_aggregated_time += (time.time() - self.batch_begin_time)
        pass

    def on_batch_end(self, mode: str, batch: int, batch_dict: Dict = None) -> None:
        logging.getLogger('Fuse').debug(f"Time for {mode} batch {batch}: {misc.get_time_delta(self.batch_begin_time)}")
        pass

    def on_train_begin(self, state: FuseManagerState) -> None:
        # update number of epochs from the manager's state:
        self.num_epochs = state.num_epochs
        self.train_begin_time = time.time()
        pass

    def on_train_end(self) -> None:
        logging.getLogger('Fuse').debug(f"Time for train: {misc.get_time_delta(self.train_begin_time)}")
        pass



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import logging
from typing import Sequence, Tuple, List, Union, Optional, Any

import numpy as np
import sklearn
from sklearn.utils import compute_sample_weight
import  matplotlib.pyplot as plt

class FuseMetricsToolBox:
    """
    Common utils for metrics
    """
    @staticmethod
    def convert_probabilities_to_class(prediction: Union[List[np.ndarray], np.ndarray], thresholds: List[Tuple] = None) -> np.array:
        """
        convert probabilities to class prediction, threshold per class
        :param prediction: either list of numpy arrays of shape [NUM_CLASSES] or numpy array of shape [N, NUM_CLASSES]
        :param thresholds: list of tuples (class_idx, threshold)
        :return: array of classes
        """
        # if no threshold specified, simply apply argmax
        if thresholds is None:
            return np.argmax(np.array(prediction), -1)

        # convert according to thresholds
        output_class = np.array([-1 for x in range(len(prediction))])
        for thr_idx, thr in enumerate(thresholds):
            class_idx = thr[0]
            class_thr = thr[1]

            # if it is the last one, set the rest of the samples with this class
            if thr_idx == (len(thresholds) - 1):
                return np.where(output_class == -1, class_idx, output_class)

            # among all the samples which not already predicted, set the ones that cross the threshold with this class
            target_idx = np.argwhere(np.logical_and(np.array(prediction)[:, class_idx] > class_thr, np.array(output_class) == -1))
            output_class[target_idx] = class_idx

        return output_class

    @staticmethod
    def find_operation_points(predictions: Sequence[np.ndarray], targets: Sequence[np.ndarray], objectives:Sequence[Tuple[int, str, float]]) -> Sequence[Tuple[float]]:
        """
        Multi class version to find operation point in classification task
        :param predictions: List of arrays of shape [NUM_CLASSES]
        :param targets: List of arrays specifying the target class per
        :param objectives: Sequence of tuples, each element including (class_index, 'sensitivity' / 'specificity', value)
        :return:
        """
        operating_point_thresholds: Sequence[Tuple[int, float]] = []
        already_predicted_indices: List[int] = []

        for op in objectives:
            cls = op[0]

            # get the predicted probability for a specific class.
            # Consider all the samples that already predicted as previous classes as probability 0
            predicted_cls = []
            for i in range(len(predictions)):
                if i in already_predicted_indices:
                    predicted_cls.append(0.0)
                else:
                    predicted_cls.append(predictions[i][cls])

            # get roc curve
            fpr, tpr, thresholds = sklearn.metrics.roc_curve(np.array(targets), np.array(predicted_cls), pos_label=cls)

            # find operating point given the objective and target
            objective_str = op[1]
            assert objective_str in ["sensitivity", "specificity", None]
            if objective_str is not None:
                target = op[2]
                if objective_str == "sensitivity":
                    objective = tpr - target
                elif objective_str == "specificity":
                    objective = 1 - fpr - target
                objective[objective < 0.0] = 1.0
                operating_point_index = np.max(np.where(objective.min() == objective))
                threshold = thresholds[operating_point_index]

                # mark all the samples that cross this as threshold as already predicted.
                predicted = [i for i, val in enumerate(thresholds) if val > threshold]
                already_predicted_indices.extend(predicted)

                # add this operating point to the returned list
                operating_point_thresholds.append((cls, threshold))
            else:
                # in  a case of None objective all the rest of the samples prediction will be that specific class
                operating_point_thresholds.append((cls, 0.0))

        return operating_point_thresholds

    @staticmethod
    def confusion_metrics(prediction: np.array, cls_ind:int, target:np.array, metrics:Sequence[str] = tuple(), class_weight: Any = None)->Tuple[float]:
        """
        Compute confusion metrics
        :param prediction: class predictions
        :param cls_ind: the class to compute the metrics for in one vs all manner
        :param target: the target classes
        :param metrics: required metrics names, options: 'sensitivity', 'recall', 'tpr', 'specificity',  'selectivity', 'npr', 'precision', 'ppv', 'f1'
        :param class_weight: use None to avoid form class weight or see available options in sklearn.utils.compute_sample_weight()
        :return: dictionary, including the computed values for the required metrics
        """
        if class_weight is not None:
            sample_weight = compute_sample_weight(class_weight=class_weight, y=target)
        else:
            sample_weight = np.ones_like(target)

        class_target_t = np.where(target == cls_ind, 1, 0)
        class_pred_t = np.where(prediction == cls_ind, 1, 0)
        res = {}
        res['tp'] = (np.logical_and(class_target_t, class_pred_t)*sample_weight).sum()
        res['fn'] = (np.logical_and(class_target_t, np.logical_not(class_pred_t))*sample_weight).sum()
        res['fp'] = (np.logical_and(np.logical_not(class_target_t), class_pred_t)*sample_weight).sum()
        res['tn'] = (np.logical_and(np.logical_not(class_target_t), np.logical_not(class_pred_t))*sample_weight).sum()

        for metric in metrics:
            if metric in ['sensitivity', 'recall', 'tpr']:
                res[metric] = res['tp'] / (res['tp'] + res['fn'])
            elif metric in ['specificity', 'selectivity', 'tnr']:
                res[metric] = res['tn'] / (res['tn'] + res['fp'])
            elif metric in ['precision', 'ppv']:
                res[metric] = res['tp'] / (res['tp'] + res['fp'])
            elif metric in ['f1']:
                res[metric] = 2 * res['tp'] / (2 * res['tp'] + res['fp'] + res['fn'])
            else:
                msg = f'unknown metric {metric}'
                logging.getLogger('Fuse').error(msg)
                raise Exception(msg)

        return res

    @staticmethod
    def get_class_names(num_classes: int, class_names: Optional[Sequence[str]] = None) -> List[str]:
        """
        Get class_names either automatically generated by the given the number of classes or keep the ones provided by the user
        :param num_classes: number of classes
        :param class_names: Optional, option for the user to specify the class names
        :return: list of class names
        """
        if class_names is None:
            num_classes = num_classes
            res_class_names = ['class_' + str(i) for i in range(num_classes)]
        else:
            res_class_names = class_names

        return res_class_names

    @staticmethod
    def roc_curve(predictions: Sequence[np.ndarray], targets: Sequence[np.ndarray], samples_weight: Sequence[np.ndarray],  class_names: str, output_filename: str) -> None:
        """
        Multi class version for roc curve
        :param predictions: List of arrays of shape [NUM_CLASSES]
        :param targets: List of arrays specifying the target class per
        :return:
        """
        for cls, cls_name in enumerate(class_names):
            fpr, tpr, _ = sklearn.metrics.roc_curve(targets, np.array(predictions)[:, cls], sample_weight=samples_weight, pos_label=cls)
            auc = sklearn.metrics.auc(fpr, tpr)
            plt.plot(fpr, tpr, label=f'{cls_name}({auc:0.2f})')

        plt.title("ROC curve")
        plt.xlabel('FPR')
        plt.ylabel('TPR')
        plt.legend()
        plt.savefig(output_filename)
        plt.close()


"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

"""
Base class for caching
"""
from abc import ABC, abstractmethod
from multiprocessing import Manager
from typing import Hashable, Any, List


class FuseCacheBase(ABC):

    @abstractmethod
    def __contains__(self, key: Hashable) -> bool:
        """
        return true if key is already in cache
        :param key: any kind of hashable key
        :return: boolean. True if exist.
        """
        raise NotImplementedError

    @abstractmethod
    def __getitem__(self, key: Hashable) -> Any:
        """
        Get an item from cache. Will raise an error if key does not exist
        :param key: any kind of hashable key
        :return: the item
        """
        raise NotImplementedError

    @abstractmethod
    def __delitem__(self, key: Hashable) -> None:
        """
        Delete key. Will raise an error if key does not exist
        :param key: any kind of hashable key
        :return: None
        """
        raise NotImplementedError

    @abstractmethod
    def __setitem__(self, key: Hashable, value: Any) -> None:
        """
        Set key. Will override previous value if already exist.
        :param key: any kind of hashable key
        :param value: any kind of value to sture
        :return: None
        """
        raise NotImplementedError

    @abstractmethod
    def save(self) -> None:
        """
        Save data to cache
        :return: None
        """
        raise NotImplementedError

    @abstractmethod
    def exist(self) -> bool:
        """
        return True if cache exist and contains the samples
        """
        raise NotImplementedError

    @abstractmethod
    def reset(self) -> None:
        """
        Reset cache and delete all data
        :return: None
        """
        raise NotImplementedError

    @abstractmethod
    def get_all_keys(self, include_none: bool = False) -> List[Hashable]:
        """
        Get all keys currently cached
        :param include_none: include or filter 'none samples' which represents no samples or bad samples
        :return: List of keys
        """
        raise NotImplementedError

    def start_caching(self, manager: Manager) -> None:
        """
        start caching - the caching will be done in save().
        :param manager: multiprocessing manager to create shared data structures
        :return: None
        """
        raise NotImplementedError



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import logging
import os
from multiprocessing import Manager
from multiprocessing.pool import Pool, ThreadPool
from typing import Any, Dict, Optional, Hashable, List, Union, Tuple, Callable

import numpy as np
import torch
from pandas import DataFrame
from torch import Tensor
from tqdm import tqdm, trange

from fuse.data.augmentor.augmentor_base import FuseAugmentorBase
from fuse.data.cache.cache_base import FuseCacheBase
from fuse.data.cache.cache_files import FuseCacheFiles
from fuse.data.cache.cache_memory import FuseCacheMemory
from fuse.data.cache.cache_null import FuseCacheNull
from fuse.data.data_source.data_source_base import FuseDataSourceBase
from fuse.data.dataset.dataset_base import FuseDatasetBase
from fuse.data.processor.processor_base import FuseProcessorBase
from fuse.data.visualizer.visualizer_base import FuseVisualizerBase
from fuse.utils.utils_debug import FuseUtilsDebug
from fuse.utils.utils_hierarchical_dict import FuseUtilsHierarchicalDict
from fuse.utils.utils_logger import log_object_input_state
from fuse.utils.utils_misc import get_pretty_dataframe, FuseUtilsMisc


class FuseDatasetGenerator(FuseDatasetBase):
    """
    Fuse Dataset Generator
    Used when it's more convient to generate sevral samples at once
    """

    #### CONSTRUCTOR
    def __init__(self, data_source: FuseDataSourceBase, processor: FuseProcessorBase,
                 cache_dest: Optional[Union[str, int]] = None, augmentor: Optional[FuseAugmentorBase] = None,
                 visualizer: Optional[FuseVisualizerBase] = None, post_processing_func=None,
                 statistic_keys: Optional[List[str]] = None,
                 filter_keys: Optional[List[str]] = None):
        """
        :param data_source: objects provides the list of object description
        :param processor: data generator
        :param cache_dest: Optional, path to save caching
        :param augmentor: Optional, object that perform the augmentation
        :param visualizer: Optional, object that visualize the data
        :param post_processing_func: callback that allows to dynamically modify the data.
               Called as last step (after augmentation)
        :param statistic_keys: Optional. list of statistic keys to output in default self.summary() implementation
        :param filter_keys: Optional. list of keys to remove from the sample dictionary when getting an item
        """
        # log object input state
        log_object_input_state(self, locals())

        super().__init__()

        # store input params
        self.cache_dest = cache_dest
        self.augmentor = augmentor
        self.visualizer = visualizer
        self.processor = processor
        self.data_source = data_source
        self.post_processing_func = post_processing_func
        self.statistic_keys = statistic_keys or []
        self.filter_keys = filter_keys or []
        # initial values
        # map sample running index to sample description (mush be hashable)
        self.subsets_description = []

        # create default cache for now - the cache will be created and loaded in create()
        self.cache: FuseCacheBase = FuseCacheMemory()
        # create dummy cache
        # self.cache_fields is used to store specific fields of the sample -
        # used to optimize the running time of dataset.get(key=<key name>, use_cache=True)
        self.cache_fields: FuseCacheBase = FuseCacheNull()

        # debug modes - read configuration
        self.sample_stages_debug = FuseUtilsDebug().get_setting('dataset_sample_stages_info') != 'default'
        self.sample_user_debug = FuseUtilsDebug().get_setting('dataset_user') != 'default'

    def create(self, reset_cache: bool = False,
               num_workers: int = 16, worker_init_func: Callable = None, worker_init_args: Any = None,
               override_datasource: Optional[FuseDataSourceBase] = None, override_cache_dest: Optional[str] = None,
               pool_type: str = 'process') -> None:

        """
        Create the data set, including loading sample descriptions and caching
        :param reset_cache: if False and cache_all is True, will use load caching instead of re creating it.
        :param num_workers: number of workers used for caching
        :param worker_init_func: process initialization function (multi processing mode)
        :param worker_init_args: worker init function arguments
        :param override_datasource: might be used to change the data source
        :param override_cache_dest: might be user to change the cache destination
        :param pool_type: multiprocess pooling type, can be either 'thread' (for ThreadPool) or 'process' (for 'Pool', default).
        :return: None
        """
        # debug - override num workers
        override_num_workers = FuseUtilsDebug().get_setting('dataset_override_num_workers')
        if override_num_workers != 'default':
            num_workers = override_num_workers
            logging.getLogger('Fuse').info(f'Dataset - debug mode - override num workers to {override_num_workers}', {'color': 'red'})

        assert pool_type in ['thread', 'process'], f'Invalid pool_type: {pool_type}. Multiprocessing pooling type can be either "thread" or "process"'
        self.pool_type = pool_type

        # override data source if required
        if override_datasource is not None:
            self.data_source = override_datasource
        # override destination cache if required
        if override_cache_dest is not None:
            self.cache_dest = override_cache_dest
        # extract list of sample description
        self.subsets_description = self.data_source.get_samples_description()

        # debug - override number of samples
        dataset_override_num_samples = FuseUtilsDebug().get_setting('dataset_override_num_samples')
        if dataset_override_num_samples != 'default':
            self.subsets_description = self.subsets_description[:dataset_override_num_samples]
            logging.getLogger('Fuse').info(f'Dataset - debug mode - override num samples to {dataset_override_num_samples}', {'color': 'red'})

        # cache object
        if isinstance(self.cache_dest, str) and self.cache_dest == 'memory':
            self.cache: FuseCacheBase = FuseCacheMemory()
        elif isinstance(self.cache_dest, str):
            self.cache: FuseCacheBase = FuseCacheFiles(self.cache_dest, reset_cache)

        # cache samples if required
        if not isinstance(self.cache, FuseCacheNull):
            self.cache_all_samples(num_workers=num_workers, worker_init_func=worker_init_func, worker_init_args=worker_init_args)

            # update descriptors
            all_descriptors = self.subsets_description
            cached_descriptors = self.cache.get_all_keys()
            self.samples_description = sorted([desc for desc in cached_descriptors if desc[0] in all_descriptors])

        self.sample_descriptor_to_index = {v: k for k, v in enumerate(self.samples_description)}
    #### ITERATE AND GET DATA
    def __len__(self):
        return len(self.samples_description)

    def get(self, index: Optional[Union[int, Hashable]], key: Optional[str] = None, use_cache: bool = True) -> Any:
        """
        Get input, ground truth or metadata of a sample.

        :param index: the index of the item, if None will return all items.
                      If not an int or None, will assume that imdex is sample descriptor
        :param key: string representing the exact information required. If None, will return all sample
        :param use_cache: if true, will try to reload the sample from caching mechanism
        :return: the required info
        """
        if index is not None and not isinstance(index, int):
            # get sample giving sample descriptor
            # assume index is sample description
            index = self.samples_description.index(index)

        # if key not specified return the all sample
        if key is None:
            assert index != -1, 'get all samples is not supported when key = None'
            return self.getitem(index)

        assert use_cache == True, f'{type(self)} support only use_cache=True'

        if index is None:
            # return all samples
            values = []
            for index in trange(len(self)):
                # first look for the specific file inside the cache
                desc_field = (self.samples_description[index], key)
                if desc_field in self.cache_fields:
                    values.append(self.cache_fields[desc_field])
                else:
                    # if not found get the all sample and then extract the specified field
                    values.append(FuseUtilsHierarchicalDict.get(self.getitem(index, apply_augmentation=False), key))
            return values
        else:
            # return single sample
            # first look for the specific file inside the cache
            desc_field = (self.samples_description[index], key)
            if desc_field in self.cache_fields:
                return self.cache_fields[desc_field]
            else:
                # if not found get the all sample and then extract the specified field
                return FuseUtilsHierarchicalDict.get(self.getitem(index, apply_augmentation=False), key)

    def __getitem__(self, index: int) -> Any:
        """
        Get sample, read it from cache if possible, apply augmentation and post processing
        :param index: sample index
        :return: the required sample after augmentation
        """
        sample_stages_debug = self.sample_stages_debug
        return self.getitem(index, sample_stages_debug=sample_stages_debug)

    def getitem(self, index: int, apply_augmentation: bool = True, apply_post_processing: bool = True, sample_stages_debug: bool = False) -> Any:
        """
        Get sample, read it from cache if possible
        :param index: sample index
        :param apply_augmentation: if true, will apply augmentation
        :param apply_post_processing: If true, will apply post processing
        :param sample_stages_debug: True will log the sample dict after each stage
        :return: the required sample after augmentation
        """

        # load from cache
        sample_desc = self.samples_description[index]
        sample = self.cache[sample_desc]

        # filter some of the keys if required
        if self.filter_keys is not None:
            for key in self.filter_keys:
                try:
                    FuseUtilsHierarchicalDict.pop(sample, key)
                except KeyError:
                    pass

        # debug mode - print original sample before augmentation and before post processing
        if sample_stages_debug:
            lgr = logging.getLogger('Fuse')
            sample_str = FuseUtilsMisc.batch_dict_to_string(sample)
            lgr.info(f'Dataset - original sample:', {'color': 'green', 'attrs': 'bold'})
            lgr.info(f'{sample_str}', {'color': 'green'})
            # one time print
            self.sample_stages_debug = False

        # apply augmentation if enabled
        if self.augmentor is not None and apply_augmentation:
            sample = self.augmentor(sample)

            # debug mode - print sample after augmentation
            if sample_stages_debug:
                lgr = logging.getLogger('Fuse')
                sample_str = FuseUtilsMisc.batch_dict_to_string(sample)
                lgr.info(f'Dataset - augmented sample:', {'color': 'green', 'attrs': 'bold'})
                lgr.info(f'{sample_str}', {'color': 'green'})

        # apply post processing
        if self.post_processing_func is not None and apply_post_processing:
            self.post_processing_func(sample)

            # debug mode - print sample after post processing
            if sample_stages_debug:
                lgr = logging.getLogger('Fuse')
                sample_str = FuseUtilsMisc.batch_dict_to_string(sample)
                lgr.info(f'Dataset - post processed sample:', {'color': 'green', 'attrs': 'bold'})
                lgr.info(f'{sample_str}', {'color': 'green'})

        return sample

    #### BATCHING
    def collate_fn(self, samples: List[Dict], avoid_stack_keys: Tuple = tuple()) -> Dict:
        """
        collate list of samples into batch_dict
        :param samples: list of samples
        :param avoid_stack_keys: list of keys to just collect to a list and avoid stack operation
        :return: batch_dict
        """
        batch_dict = {}
        keys = FuseUtilsHierarchicalDict.get_all_keys(samples[0])
        for key in keys:
            try:
                collected_value = [FuseUtilsHierarchicalDict.get(sample, key) for sample in samples if sample is not None]
                if key in avoid_stack_keys:
                    FuseUtilsHierarchicalDict.set(batch_dict, key, collected_value)
                elif isinstance(collected_value[0], Tensor):
                    FuseUtilsHierarchicalDict.set(batch_dict, key, torch.stack(collected_value))
                elif isinstance(collected_value[0], np.ndarray):
                    FuseUtilsHierarchicalDict.set(batch_dict, key, np.stack(collected_value))
                else:
                    FuseUtilsHierarchicalDict.set(batch_dict, key, collected_value)
            except:
                logging.getLogger('Fuse').error(f'Failed to collect key {key}')
                raise

        return batch_dict

    #### CACHING
    def cache_all_samples(self, num_workers: int = 16, worker_init_func: Callable = None, worker_init_args: Any = None) -> None:
        """
        Cache all data
        :param num_workers: num of workers used to cache the samples
        :param worker_init_func: process initialization function (multi processing mode)
        :param worker_init_args: worker init function arguments
        :return: None
        """
        lgr = logging.getLogger('Fuse')

        # check if cache is required
        all_descriptors = set([(subset_desc, 0) for subset_desc in self.subsets_description])
        cached_descriptors = set(self.cache.get_all_keys(include_none=True))
        descriptors_to_cache = all_descriptors - cached_descriptors

        if len(descriptors_to_cache) != 0:
            # multi process cache
            lgr.info(f'FuseDatasetGenerator: caching {len(descriptors_to_cache)} out of {len(all_descriptors)}')
            with  Manager() as manager:
                # change cache mode - to caching (writing)
                self.cache.start_caching(manager)

                # multi process cache
                if num_workers > 0:
                    the_pool = ThreadPool if self.pool_type == 'thread' else Pool
                    pool = the_pool(processes=num_workers, initializer=worker_init_func, initargs=worker_init_args)
                    for _ in tqdm(pool.imap_unordered(func=self._cache_subset,
                                                      iterable=[(self.processor, subset_desc[0], self.cache) for subset_desc in descriptors_to_cache]),
                                  total=len(descriptors_to_cache), smoothing=0.1):
                        pass
                    pool.close()
                    pool.join()
                else:
                    for subset_desc in tqdm(descriptors_to_cache):
                        self._cache_subset((self.processor, subset_desc[0], self.cache))

                # save and move back to read mode
                self.cache.save()
                lgr.info('FuseDatasetGenerator: caching done')
        else:
            lgr.info('FuseDatasetGenerator: all samples are already cached')

    @staticmethod
    def _cache_subset(args: Tuple) -> None:
        """
        Store in cache single sample
        :param args: tuple of processor and subset descriptor
        :return: None
        """
        processor, subset_desc, cache = args
        samples = processor(subset_desc)
        if not isinstance(samples, List):
            samples = [samples]
        if samples:
            for sample_index, sample_data in enumerate(samples):

                assert isinstance(sample_data, dict), f'expecting sample_data to be dictionary, got {type(sample_data)}'
                sample_data = sample_data.copy()

                sample = {'data': sample_data}
                sample_data['descriptor'] = (subset_desc, sample_index)
                cache[sample_data['descriptor']] = sample
        else:
            # no samples extracted mark it as an invalid descriptor
            cache[(subset_desc, 0)] = None

    def cache_sample_fields(self, fields: List[str], reset_cache: bool = False, num_workers: int = 8, cache_dest: Optional[str] = None) -> None:
        """
        Cache specific fields (keys in batch_dict)
        Used to optimize the running time of of dataset.get(key=<key name>, use_cache=True)
        :param fields: list of keys in batch_dict
        :param reset_cache: If True will reset cache first
        :param num_workers: num workers used for caching
        :param cache_dest: path to cache dir
        :return: None
        """
        lgr = logging.getLogger('Fuse')

        # debug - override num workers
        override_num_workers = FuseUtilsDebug().get_setting('dataset_override_num_workers')
        if override_num_workers != 'default':
            num_workers = override_num_workers
            lgr.info(f'Dataset - debug mode - override num workers to {override_num_workers}', {'color': 'red'})

        if cache_dest is None:
            cache_dest = os.path.join(self.cache_dest, 'fields')

        # create cache field object upon request
        if isinstance(self.cache_fields, FuseCacheNull):
            # cache object
            if isinstance(cache_dest, str) and cache_dest == 'memory':
                self.cache_fields: FuseCacheBase = FuseCacheMemory()
            elif isinstance(cache_dest, str):
                self.cache_fields: FuseCacheBase = FuseCacheFiles(cache_dest, reset_cache, single_file=True)

        # get list of desc to cache
        desc_list = self.samples_description
        desc_field_list = set([(desc, field) for desc in desc_list for field in fields])
        cached_desc_field = set(self.cache_fields.get_all_keys(include_none=True))
        desc_field_to_cache = desc_field_list - cached_desc_field
        desc_to_cache = set([desc_field[0] for desc_field in desc_field_to_cache])

        # multi thread caching
        if len(desc_to_cache) != 0:
            lgr.info(f'FuseDatasetGenerator: samples fields - caching {len(desc_to_cache)} out of {len(desc_list)}')
            if num_workers > 0:
                with Manager() as manager:
                    self.cache_fields.start_caching(manager)
                    pool = Pool(processes=num_workers)
                    for _ in tqdm(pool.imap_unordered(func=self._cache_sample_fields,
                                                      iterable=[(desc, fields) for desc in desc_to_cache]),
                                  total=len(desc_to_cache), smoothing=0.1):
                        pass
                    pool.close()
                    pool.join()
                    self.cache_fields.save()
            else:
                self.cache_fields.start_caching(None)
                for desc in tqdm(desc_to_cache):
                    self._cache_sample_fields((desc, fields))
                self.cache_fields.save()
        else:
            lgr.info('FuseDatasetGenerator: all samples fields are already cached')

    def _cache_sample_fields(self, args):
        # decode args
        desc, fields = args
        index = self.samples_description.index(desc)
        sample = self.getitem(index, apply_augmentation=False)
        for field in fields:
            # create field desc and save it in cache
            desc_field = (desc, field)
            if desc_field not in self.cache_fields:
                value = FuseUtilsHierarchicalDict.get(sample, field)
                self.cache_fields[desc_field] = value

    #### Filtering
    def filter(self, key: str, values: List[Any]) -> None:
        """
        Filter sample if batch_dict[key] in values
        :param key: key in batch_dict
        :param values: list of values to filter
        :return: None
        """
        lgr = logging.getLogger('Fuse')
        lgr.info(f'DatasetGenerator: filtering key {key}, values {values}')
        new_samples_desc = []
        for index, desc in tqdm(enumerate(self.samples_description), total=len(self.samples_description)):
            value = self.get(index, key, use_cache=True)
            if value not in values:
                new_samples_desc.append(desc)

        self.samples_description = new_samples_desc



    #### VISUALISE
    def visualize(self, index: Optional[int] = None, descriptor: Optional[Hashable] = None, block: bool = True):
        """
        visualize sample
        :param index: sample index, only one of index/descriptor can be provided
        :param descriptor: descriptor of a sample , only one of index/descriptor can be provided
        :param block: set to False if the process should not be blocked until the plot will be closed
        :return: None
        """
        assert (index is not None) ^ (descriptor is not None), "visualize method must get one and one only of an index or a descriptor"
        lgr = logging.getLogger('Fuse')
        if descriptor is not None:
            index = self.sample_descriptor_to_index[descriptor]

        if self.visualizer is None:
            lgr.warning('Cannot visualize - visualizer was not provided')
            return

        batch_dict = self.getitem(index)

        self.visualizer.visualize(batch_dict, block)

    def visualize_augmentation(self, index: Optional[int] = None, descriptor: Optional[Hashable] = None, block: bool = True):
        """
        visualize augmentation of a sample
        :param index: sample index, only one of index/descriptor can be provided
        :param descriptor: descriptor of a sample, only one of index/descriptor can be provided
        :param block: set to False if the process should not be blocked until the plot will be closed
        :return: None
        """

        assert (index is not None) ^ (descriptor is not None), "visualize method must get one and one only of an index or a descriptor"

        lgr = logging.getLogger('Fuse')
        if descriptor is not None:
            index = self.sample_descriptor_to_index[descriptor]
        if self.visualizer is None:
            lgr.warning('Cannot visualize - visualizer was not provided')
            return
        batch_dict = self.getitem(index, apply_augmentation=False)
        batch_dict_aug = self.getitem(index)

        self.visualizer.visualize_aug(batch_dict, batch_dict_aug, block)

    # save and load dataset
    def get_instance_to_save(self, mode: FuseDatasetBase.SaveMode) -> FuseDatasetBase:
        """
        See base class
        """

        # prepare data to save
        if mode == FuseDatasetBase.SaveMode.INFERENCE:
            dataset = FuseDatasetGenerator(data_source=None,
                                           processor=self.processor,
                                           augmentor=self.augmentor,
                                           post_processing_func=self.post_processing_func
                                           )
        elif mode == FuseDatasetBase.SaveMode.TRAINING:
            dataset = FuseDatasetGenerator(data_source=self.data_source,
                                           processor=self.processor,
                                           augmentor=self.augmentor,
                                           post_processing_func=self.post_processing_func,
                                           visualizer=self.visualizer)
        else:
            raise Exception(f'Unexpected SaveMode {mode}')

        return dataset

    # misc
    def summary(self, statistic_keys: Optional[List[str]] = None) -> str:
        """
        Returns a data summary.
        Should be called after create()
        :param statistic_keys: Optional. list of keys to output statistics about.
                        When None (default), self.statistic_keys are output.
        :return: str
        """
        statistic_keys_to_use = statistic_keys if statistic_keys is not None else self.statistic_keys
        sum = \
            f'Class = {self.__class__}\n'
        sum += \
            f'Processor:\n' \
                f'-----------------\n' \
                f'{self.processor}\n'
        sum += \
            f'Cache destination:\n' \
                f'------------------\n' \
                f'{self.cache_dest}\n'
        sum += \
            f'Augmentor:\n' \
                f'----------\n' \
                f'{self.augmentor.summary() if self.augmentor is not None else None}\n'
        sum += \
            f'Sample keys:\n' \
                f'------------\n' \
                f'{FuseUtilsHierarchicalDict.get_all_keys(self.getitem(0)) if self.data_source is not None else None}\n'
        if len(statistic_keys_to_use) > 0:
            for key in statistic_keys_to_use:
                values = self.get(index=None, key=key, use_cache=True)
                # convert to int - maybe we will need to supporty additional types one day
                values = [int(value) for value in values]
                df = DataFrame(data=values,
                               columns=[key])
                stat_df = DataFrame()
                stat_df['Value'] = df[key].value_counts().index
                stat_df['Count'] = df[key].value_counts().values
                stat_df['Percent'] = df[key].value_counts(normalize=True).values * 100
                sum += \
                    f'\n{key} Statistics:\n' + \
                    f'{get_pretty_dataframe(stat_df)}'
        return sum



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

"""
Data source base
"""
from abc import ABC, abstractmethod


class FuseDataSourceBase(ABC):

    @abstractmethod
    def get_samples_description(self):
        """
        :return: list of samples description
        """
        raise NotImplementedError

    @abstractmethod
    def summary(self) -> str:
        """
        String summary of the object
        """
        raise NotImplementedError



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""
import numpy as np
import skimage
import skimage.io as io
import skimage.transform as transform
import torch

import logging
import traceback
from typing import Optional, Tuple

from fuse.data.processor.processor_base import FuseProcessorBase


class FuseSkinInputProcessor(FuseProcessorBase):
    def __init__(self,
                 input_data: str,
                 normalized_target_range: Tuple = (-1, 1),
                 resize_to: Optional[Tuple] = (299, 299),
                 padding: Optional[Tuple] = (0, 0),
                 ):
        """
        Create Input processor
        :param input_data:              path to images
        :param normalized_target_range: range for image normalization
        :param resize_to:               Optional, new size of input images, keeping proportions
        :param padding:                 Optional, padding size
        """

        self.input_data = input_data
        self.normalized_target_range = normalized_target_range
        self.resize_to = np.subtract(resize_to, (2*padding[0], 2*padding[1]))
        self.padding = padding

    def __call__(self,
                 inner_image_desc,
                 *args, **kwargs):
        try:
            img_path = self.input_data + str(inner_image_desc) + '.jpg'

            # read image
            inner_image = skimage.io.imread(img_path)

            # convert to numpy
            inner_image = np.asarray(inner_image)
            
            # normalize
            inner_image = normalize_to_range(inner_image, range=self.normalized_target_range)

            # resize
            inner_image_height, inner_image_width = inner_image.shape[0], inner_image.shape[1]
            
            if self.resize_to is not None:
                if inner_image_height > self.resize_to[0]:
                    h_ratio = self.resize_to[0] / inner_image_height
                else:
                    h_ratio = 1
                if inner_image_width > self.resize_to[1]:
                    w_ratio = self.resize_to[1] / inner_image_width
                else:
                    w_ratio = 1

                resize_ratio = min(h_ratio, w_ratio)
                if resize_ratio != 1:
                    inner_image = skimage.transform.resize(inner_image,
                                                           output_shape=(int(inner_image_height * resize_ratio),
                                                                         int(inner_image_width * resize_ratio)),
                                                           mode='reflect',
                                                           anti_aliasing=True
                                                           )

            # padding
            if self.padding is not None:
                # "Pad" around inner image
                inner_image = inner_image.astype('float32')

                inner_image_height, inner_image_width = inner_image.shape[0], inner_image.shape[1]
                inner_image[0:inner_image_height, 0] = 0
                inner_image[0:inner_image_height, inner_image_width-1] = 0
                inner_image[0, 0:inner_image_width] = 0
                inner_image[inner_image_height-1, 0:inner_image_width] = 0

                if self.normalized_target_range is None:
                    pad_value = 0
                else:
                    pad_value = self.normalized_target_range[0]

                image = pad_image(inner_image, outer_height=self.resize_to[0] + 2*self.padding[0], outer_width=self.resize_to[1] + 2*self.padding[1], pad_value=pad_value)

            else:
                image = inner_image

            # convert image from shape (H x W x C) to shape (C x H x W) with C=3
            image = np.moveaxis(image, -1, 0)

            # numpy to tensor
            sample = torch.from_numpy(image)

        except Exception as e:
            lgr = logging.getLogger('Fuse')
            track = traceback.format_exc()
            lgr.error(e)
            lgr.error(track)
            return None

        return sample


def normalize_to_range(input_image: np.ndarray, range: Tuple = (-1.0, 1.0)):
    """
    Scales tensor to range
    @param input_image: image of shape (H x W x C)
    @param range:       bounds for normalization
    @return:            normalized image
    """
    max_val = input_image.max()
    min_val = input_image.min()
    if min_val == max_val == 0:
        return input_image
    input_image = input_image - min_val
    input_image = input_image / (max_val - min_val)
    input_image = input_image * (range[1] - range[0])
    input_image = input_image + range[0]
    return input_image


def pad_image(image: np.ndarray, outer_height: int, outer_width: int, pad_value: Tuple):
    """
    Pastes input image in the middle of a larger one
    @param image:        image of shape (H x W x C)
    @param outer_height: final outer height
    @param outer_width:  final outer width
    @param pad_value:    value for padding around inner image
    @return:             padded image
    """
    inner_height, inner_width = image.shape[0], image.shape[1]
    h_offset = int((outer_height - inner_height) / 2.0)
    w_offset = int((outer_width - inner_width) / 2.0)
    outer_image = np.ones((outer_height, outer_width, 3), dtype=image.dtype) * pad_value
    outer_image[h_offset:h_offset + inner_height, w_offset:w_offset + inner_width, :] = image

    return outer_image



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import logging
import os

import torch
import torch.nn.functional as F
import torch.optim as optim
import torchvision
import torchvision.models as models
from torch.utils.data.dataloader import DataLoader
from torchvision import transforms

from fuse.analyzer.analyzer_default import FuseAnalyzerDefault
from fuse.data.dataset.dataset_wrapper import FuseDatasetWrapper
from fuse.data.sampler.sampler_balanced_batch import FuseSamplerBalancedBatch
from fuse.losses.loss_default import FuseLossDefault
from fuse.managers.callbacks.callback_metric_statistics import FuseMetricStatisticsCallback
from fuse.managers.callbacks.callback_tensorboard import FuseTensorboardCallback
from fuse.managers.callbacks.callback_time_statistics import FuseTimeStatisticsCallback
from fuse.managers.manager_default import FuseManagerDefault
from fuse.metrics.classification.metric_accuracy import FuseMetricAccuracy
from fuse.metrics.classification.metric_auc import FuseMetricAUC
from fuse.metrics.classification.metric_roc_curve import FuseMetricROCCurve
from fuse.models.model_wrapper import FuseModelWrapper
from fuse.utils.utils_debug import FuseUtilsDebug
from fuse.utils.utils_gpu import FuseUtilsGPU
from fuse.utils.utils_logger import fuse_logger_start

###########################################################################################################
# Fuse
###########################################################################################################
##########################################
# Debug modes
##########################################
mode = 'default'  # Options: 'default', 'fast', 'debug', 'verbose', 'user'. See details in FuseUtilsDebug
debug = FuseUtilsDebug(mode)

##########################################
# Output Paths
##########################################
ROOT = 'examples' # TODO: fill path here
PATHS = {'model_dir': os.path.join(ROOT, 'mnist/model_dir'),
         'force_reset_model_dir': True,  # If True will reset model dir automatically - otherwise will prompt 'are you sure' message.
         'cache_dir': os.path.join(ROOT, 'mnist/cache_dir'),
         'inference_dir': os.path.join(ROOT, 'mnist/infer_dir'),
         'analyze_dir': os.path.join(ROOT, 'mnist/analyze_dir')}

##########################################
# Train Common Params
##########################################
# ============
# Data
# ============
TRAIN_COMMON_PARAMS = {}
TRAIN_COMMON_PARAMS['data.batch_size'] = 30
TRAIN_COMMON_PARAMS['data.train_num_workers'] = 8
TRAIN_COMMON_PARAMS['data.validation_num_workers'] = 8

# ===============
# Manager - Train
# ===============
TRAIN_COMMON_PARAMS['manager.train_params'] = {
    'device': 'cuda', 
    'num_epochs': 5,
    'virtual_batch_size': 1,  # number of batches in one virtual batch
    'start_saving_epochs': 10,  # first epoch to start saving checkpoints from
    'gap_between_saving_epochs': 5,  # number of epochs between saved checkpoint
}
TRAIN_COMMON_PARAMS['manager.best_epoch_source'] = {
    'source': 'metrics.accuracy',  # can be any key from 'epoch_results'
    'optimization': 'max',  # can be either min/max
    'on_equal_values': 'better',
    # can be either better/worse - whether to consider best epoch when values are equal
}
TRAIN_COMMON_PARAMS['manager.learning_rate'] = 1e-4
TRAIN_COMMON_PARAMS['manager.weight_decay'] = 0.001
TRAIN_COMMON_PARAMS['manager.resume_checkpoint_filename'] = None  # if not None, will try to load the checkpoint


def perform_softmax(output):
    if isinstance(output, torch.Tensor):  # validation
        logits = output
    else:  # train
        logits = output.logits
    cls_preds = F.softmax(logits, dim=1)
    return logits, cls_preds


#################################
# Train Template
#################################
def run_train(paths: dict, train_params: dict):
    # ==============================================================================
    # Logger
    # ==============================================================================
    fuse_logger_start(output_path=paths['model_dir'], console_verbose_level=logging.INFO)
    lgr = logging.getLogger('Fuse')
    lgr.info('Fuse Train', {'attrs': ['bold', 'underline']})

    lgr.info(f'model_dir={paths["model_dir"]}', {'color': 'magenta'})
    lgr.info(f'cache_dir={paths["cache_dir"]}', {'color': 'magenta'})

    # ==============================================================================
    # Data
    # ==============================================================================
    # Train Data
    lgr.info(f'Train Data:', {'attrs': 'bold'})

    transform = transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize((0.1307,), (0.3081,))
    ])
    # Create dataset
    torch_train_dataset = torchvision.datasets.MNIST(paths['cache_dir'], download=True, train=True, transform=transform)
    # wrapping torch dataset
    # FIXME: support also using torch dataset directly
    train_dataset = FuseDatasetWrapper(name='train', dataset=torch_train_dataset, mapping=('image', 'label'))
    train_dataset.create()
    lgr.info(f'- Create sampler:')
    sampler = FuseSamplerBalancedBatch(dataset=train_dataset,
                                       balanced_class_name='data.label',
                                       num_balanced_classes=10,
                                       batch_size=train_params['data.batch_size'],
                                       balanced_class_weights=None)
    lgr.info(f'- Create sampler: Done')

    # Create dataloader
    train_dataloader = DataLoader(dataset=train_dataset, batch_sampler=sampler, num_workers=train_params['data.train_num_workers'])
    lgr.info(f'Train Data: Done', {'attrs': 'bold'})

    ## Validation data
    lgr.info(f'Validation Data:', {'attrs': 'bold'})
    # Create dataset
    torch_validation_dataset = torchvision.datasets.MNIST(paths['cache_dir'], download=True, train=False, transform=transform)
    # wrapping torch dataset
    validation_dataset = FuseDatasetWrapper(name='validation', dataset=torch_validation_dataset, mapping=('image', 'label'))
    validation_dataset.create()

    # dataloader
    validation_dataloader = DataLoader(dataset=validation_dataset, batch_size=train_params['data.batch_size'],
                                       num_workers=train_params['data.validation_num_workers'])
    lgr.info(f'Validation Data: Done', {'attrs': 'bold'})

    # ==============================================================================
    # Model
    # ==============================================================================
    lgr.info('Model:', {'attrs': 'bold'})

    torch_model = models.resnet18(num_classes=10)
    # modify conv1 to support single channel image
    torch_model.conv1 = torch.nn.Conv2d(1, 64, kernel_size=7, stride=2, padding=3, bias=False)
    # use adaptive avg pooling to support mnist low resolution images
    torch_model.avgpool = torch.nn.AdaptiveAvgPool2d(1)

    model = FuseModelWrapper(model=torch_model,
                             model_inputs=['data.image'],
                             post_forward_processing_function=perform_softmax,
                             model_outputs=['logits.classification', 'output.classification']
                             )

    lgr.info('Model: Done', {'attrs': 'bold'})

    # ====================================================================================
    #  Loss
    # ====================================================================================
    losses = {
        'cls_loss': FuseLossDefault(pred_name='model.logits.classification', target_name='data.label', callable=F.cross_entropy, weight=1.0),
    }

    # ====================================================================================
    # Metrics
    # ====================================================================================
    metrics = {
        'accuracy': FuseMetricAccuracy(pred_name='model.output.classification', target_name='data.label')
    }

    # =====================================================================================
    #  Callbacks
    # =====================================================================================
    callbacks = [
        # default callbacks
        FuseTensorboardCallback(model_dir=paths['model_dir']),  # save statistics for tensorboard
        FuseMetricStatisticsCallback(output_path=paths['model_dir'] + "/metrics.csv"),  # save statistics a csv file
        FuseTimeStatisticsCallback(num_epochs=train_params['manager.train_params']['num_epochs'], load_expected_part=0.1)  # time profiler
    ]

    # =====================================================================================
    #  Manager - Train
    # =====================================================================================
    lgr.info('Train:', {'attrs': 'bold'})

    # create optimizer
    optimizer = optim.Adam(model.parameters(), lr=train_params['manager.learning_rate'], weight_decay=train_params['manager.weight_decay'])

    # create learning scheduler
    scheduler = optim.lr_scheduler.ReduceLROnPlateau(optimizer)

    # train from scratch
    manager = FuseManagerDefault(output_model_dir=paths['model_dir'], force_reset=paths['force_reset_model_dir'])
    # Providing the objects required for the training process.
    manager.set_objects(net=model,
                        optimizer=optimizer,
                        losses=losses,
                        metrics=metrics,
                        best_epoch_source=train_params['manager.best_epoch_source'],
                        lr_scheduler=scheduler,
                        callbacks=callbacks,
                        train_params=train_params['manager.train_params'])

    ## Continue training
    if train_params['manager.resume_checkpoint_filename'] is not None:
        # Loading the checkpoint including model weights, learning rate, and epoch_index.
        manager.load_checkpoint(checkpoint=train_params['manager.resume_checkpoint_filename'], mode='train')

    # Start training
    manager.train(train_dataloader=train_dataloader, validation_dataloader=validation_dataloader)

    lgr.info('Train: Done', {'attrs': 'bold'})


######################################
# Inference Common Params
######################################
INFER_COMMON_PARAMS = {}
INFER_COMMON_PARAMS['infer_filename'] = 'validation_set_infer.gz'
INFER_COMMON_PARAMS['checkpoint'] = 'best'  # Fuse TIP: possible values are 'best', 'last' or epoch_index.


######################################
# Inference Template
######################################
def run_infer(paths: dict, infer_common_params: dict):
    #### Logger
    fuse_logger_start(output_path=paths['inference_dir'], console_verbose_level=logging.INFO)
    lgr = logging.getLogger('Fuse')
    lgr.info('Fuse Inference', {'attrs': ['bold', 'underline']})
    lgr.info(f'infer_filename={os.path.join(paths["inference_dir"], infer_common_params["infer_filename"])}', {'color': 'magenta'})

    ## Data
    transform = transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize((0.1307,), (0.3081,))
    ])
    # Create dataset
    torch_validation_dataset = torchvision.datasets.MNIST(paths['cache_dir'], download=True, train=False, transform=transform)
    # wrapping torch dataset
    validation_dataset = FuseDatasetWrapper(name='validation', dataset=torch_validation_dataset, mapping=('image', 'label'))
    validation_dataset.create()
    # dataloader
    validation_dataloader = DataLoader(dataset=validation_dataset, collate_fn=validation_dataset.collate_fn, batch_size=2, num_workers=2)

    ## Manager for inference
    manager = FuseManagerDefault()
    output_columns = ['model.output.classification', 'data.label']
    manager.infer(data_loader=validation_dataloader,
                  input_model_dir=paths['model_dir'],
                  checkpoint=infer_common_params['checkpoint'],
                  output_columns=output_columns,
                  output_file_name=os.path.join(paths["inference_dir"], infer_common_params["infer_filename"]))


######################################
# Analyze Common Params
######################################
ANALYZE_COMMON_PARAMS = {}
ANALYZE_COMMON_PARAMS['infer_filename'] = INFER_COMMON_PARAMS['infer_filename']
ANALYZE_COMMON_PARAMS['output_filename'] = os.path.join(PATHS['analyze_dir'], 'all_metrics')


######################################
# Analyze Template
######################################
def run_analyze(paths: dict, analyze_common_params: dict):
    fuse_logger_start(output_path=None, console_verbose_level=logging.INFO)
    lgr = logging.getLogger('Fuse')
    lgr.info('Fuse Analyze', {'attrs': ['bold', 'underline']})

    # metrics
    metrics = {
        'accuracy': FuseMetricAccuracy(pred_name='model.output.classification', target_name='data.label'),
        'roc': FuseMetricROCCurve(pred_name='model.output.classification', target_name='data.label', output_filename=os.path.join(paths['inference_dir'], 'roc_curve.png')),
        'auc': FuseMetricAUC(pred_name='model.output.classification', target_name='data.label')
    }

    # create analyzer
    analyzer = FuseAnalyzerDefault()

    # run
    # FIXME: simplify analyze interface for this case
    results = analyzer.analyze(gt_processors={},
                     data_pickle_filename=os.path.join(paths["inference_dir"], analyze_common_params["infer_filename"]),
                     metrics=metrics,
                     output_filename=analyze_common_params['output_filename'])

    return results


######################################
# Run
######################################
if __name__ == "__main__":
    # allocate gpus
    NUM_GPUS = 1
    if NUM_GPUS == 0:
        TRAIN_COMMON_PARAMS['manager.train_params']['device'] = 'cpu' 
    # uncomment if you want to use specific gpus instead of automatically looking for free ones
    force_gpus = None  # [0]
    FuseUtilsGPU.choose_and_enable_multiple_gpus(NUM_GPUS, force_gpus=force_gpus)

    RUNNING_MODES = ['train', 'infer', 'analyze']  # Options: 'train', 'infer', 'analyze'
    # train
    if 'train' in RUNNING_MODES:
        run_train(paths=PATHS, train_params=TRAIN_COMMON_PARAMS)

    # infer
    if 'infer' in RUNNING_MODES:
        run_infer(paths=PATHS, infer_common_params=INFER_COMMON_PARAMS)

    # analyze
    if 'analyze' in RUNNING_MODES:
        run_analyze(paths=PATHS, analyze_common_params=ANALYZE_COMMON_PARAMS)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Hashable, List, Optional, Dict, Union
import logging
import torch
import pandas as pd
from torch import Tensor

from fuse.data.processor.processor_base import FuseProcessorBase


class FuseProcessorDataFrame(FuseProcessorBase):
    """
    Processor reading data from pickle file / dataframe object.
    Covert each row to a dictionary
    """

    def __init__(self,
                 data: Optional[pd.DataFrame] = None,
                 data_pickle_filename: Optional[str] = None,
                 sample_desc_column: Optional[str] = 'descriptor',
                 columns_to_extract: Optional[List[str]] = None,
                 rename_columns: Optional[Dict[str, str]] = None,
                 columns_to_tensor: Optional[Union[List[str], Dict[str, torch.dtype]]] = None):
        """
        :param data:  input DataFrame
        :param data_pickle_filename: path to a pickled DataFrame (possible gzipped)
        :param sample_desc_column: name of the sample descriptor column within the pickle file,
                                   if set to None.will simply use dataframe index as descriptors
        :param columns_to_extract: list of columns to extract from dataframe. When None (default) all columns are extracted
        :param rename_columns: rename columns from dataframe, when None (default) column names are kept
        :param columns_to_tensor: columns in data that should be converted into pytorch.tensor.
                        when list, all columns specified are transforms into tensors (type is decided by torch).
                        when dictionary, then each column is converted into the specified dtype.
                        When None (default) no columns are converted.
        """
        # verify input
        lgr = logging.getLogger('Fuse')
        if data is None and data_pickle_filename is None:
            msg = "Error in FuseProcessorDataFrame - need to provide either in-memory DataFrame or a path to pickled DataFrame."
            lgr.error(msg)
            raise Exception(msg)
        elif data is not None and data_pickle_filename is not None:
            msg = "Error in FuseProcessorDataFrame - need to provide either 'data' or 'data_pickle_filename' args, bot not both."
            lgr.error(msg)
            raise Exception(msg)

        # read dataframe
        if data is not None:
            self.data = data
            self.pickle_filename = 'in-memory'
        elif data_pickle_filename is not None:
            self.data = pd.read_pickle(data_pickle_filename)
            self.pickle_filename = data_pickle_filename

        # store input arguments
        self.sample_desc_column = sample_desc_column
        self.columns_to_extract = columns_to_extract
        self.columns_to_tensor = columns_to_tensor

        # extract only specified columns (in case not specified, extract all)
        if self.columns_to_extract is not None:
            self.data = self.data[self.columns_to_extract]

        # rename columns
        if rename_columns is not None:
            self.data.rename(rename_columns, axis=1, inplace=True)

        # convert to dictionary: {index -> {column -> value}}
        self.data = self.data.set_index(self.sample_desc_column)
        self.data = self.data.to_dict(orient='index')

    def __call__(self, sample_desc: Hashable):
        """
        See base class
        """
        # locate the required item
        sample_data = self.data[sample_desc].copy()

        # convert to tensor
        if self.columns_to_tensor is not None:
            if isinstance(self.columns_to_tensor, list):
                for col in self.columns_to_tensor:
                    self.convert_to_tensor(sample_data, col)
            elif isinstance(self.columns_to_tensor, dict):
                for col, tensor_dtype in self.columns_to_tensor.items():
                    self.convert_to_tensor(sample_data, col, tensor_dtype)

        return sample_data

    def get_samples_descriptors(self) -> List[Hashable]:
        """
        :return: list of descriptors dataframe index values
        """
        return list(self.data.keys())

    @staticmethod
    def convert_to_tensor(sample: dict, key: str, tensor_dtype: Optional[str] = None) -> None:
        """
        Convert value to tensor, use tensor_dtype to specify non-default type/
        :param sample: sample dictionary
        :param key: key of item in sample dict to convert
        :param tensor_dtype: Optional, None for default,.
        """
        if key not in sample:
            lgr = logging.getLogger('Fuse')
            lgr.error(f'Column {key} does not exit in dataframe, it is ignored and not converted to {tensor_dtype}')
        else:
            if isinstance(sample[key], Tensor):
                sample[key] = sample[key]
            else:
                sample[key] = torch.tensor(sample[key], dtype=tensor_dtype)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from abc import ABC
import logging
from torch import Tensor
from typing import Union, Dict, Callable, Optional

from fuse.utils.utils_hierarchical_dict import FuseUtilsHierarchicalDict


class FuseMetricBase(ABC):
    """
    Base class for Fuse metrics
    """

    def __init__(self,
                 pred_name: str,
                 target_name: str,
                 filter_func: Optional[Callable] = None,
                 use_sample_weights: Optional[bool] = False,
                 **additional_collect: str) -> None:
        """
        Base class for metrics.
        :param pred_name:           batch_dict key for predicted output (e.g., class probabilities after softmax)
        :param target_name:         batch_dict key for target (e.g., ground truth label)
        :param filter_func:         Optional - function that filters batch_dict. The function gets as input batch_dict and returns filtered batch_dict.
                                    When None (default), the entire batch dict is collected and processed.
        :param use_sample_weights:  Optional - when True (default is False), metrics are computed with sample weigths.
                                    For this, use argument 'sample_weight_name' to define the batch_dict key for the sample weight
        :param additional_collect:  args ending with '_name' will be collected into self.collected_data[].
                                    the prefix of argument is the key, and the value of the argument is the key in batch_dict to collect the data from.
                                    e.g., weights_name='data.metadata.weight':
                                            collect data from batch_dict['data.metadata.weight'] into self.collected_data['weights'] array
        """
        additional_collect.update({'pred_name': pred_name, 'target_name': target_name})
        self.key_to_collect: dict = {}
        self.collected_data: Dict[str, list] = {}
        for name_arg, key in additional_collect.items():
            if key is None:  # if user did not pass values for key
                continue
            assert name_arg.endswith('_name')
            name = name_arg[:-5]
            self.add_key_to_collect(name, key)
        if use_sample_weights:
            assert 'sample_weight' in self.key_to_collect, \
                "Metric is expected to use sample weights, but 'sample_weight_name' was not specified"
        self.use_sample_weights = use_sample_weights
        self.filter_func = filter_func
        self.reset()

    def collect(self,
                batch_dict: Dict) -> None:
        """
        Triggered after each batch - collects data for metrics. data are saved into self.collected_data dictionary.
        The dictionary contains at least the keys: predictions, targets.
        Any additional keys to collect from batch_dict can be specified by additional_collect arguments in the init()

        :return: None
        """
        if self.filter_func is not None:
            batch_dict = self.filter_func(batch_dict)

        for name, key in self.key_to_collect.items():
            values = FuseUtilsHierarchicalDict.get(batch_dict, key)
            if isinstance(values, Tensor):
                values = values.detach().cpu().numpy()
            self.collected_data[name].extend(values)

        pass

    def process(self) -> Union[float, Dict[str, float], str]:
        """
        Triggered after each epoch - calculates epoch metrics
        :return:
        """
        raise NotImplementedError

    def reset(self) -> None:
        """
        Resets collected data for metrics
        :return: None
        """
        for key in self.collected_data.keys():
            self.collected_data.update({key: []})
        pass

    # backward comparability: keep the main collected data in lists
    @property
    def epoch_preds(self):
        return self.collected_data['pred']
    @property
    def epoch_targets(self):
        return self.collected_data['target']

    def add_key_to_collect(self, name, key) -> None:
        """
        Utility function to add keys that should be collected on collect().
        Can be used after the initialization of the class, and before collect() was called (or after reset()).
        :return:  None
        """
        if name in self.key_to_collect:
            if self.key_to_collect[name] != key:
                logging.getLogger('Fuse').debug(f"Replacing key in batch_dict for {name}: was {self.key_to_collect[name]} now {key}")
        self.key_to_collect.update({name: key})
        self.collected_data.update({name: []})
        pass


"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import logging
import os
import traceback
from typing import Dict, List, Optional

import pandas as pd

from fuse.managers.callbacks.callback_base import FuseCallback
from fuse.utils.utils_file import FuseUtilsFile as file
from fuse.utils.utils_hierarchical_dict import FuseUtilsHierarchicalDict


class FuseMetricStatisticsCallback(FuseCallback):
    """
        Responsible of writing the metric results into a CSV file under output_path
        The columns are: mode, epoch, metric_name, metric_value
    """

    def __init__(self, output_path: str, metrics: Optional[List[str]] = None) -> None:
        """
        :param output_path: output csv file path
        :param metrics: list of values out of the epoch_result dictionary to save.
                        Set to, will output all the epoch results to a csv file.
        """
        super().__init__()
        self.output_path = output_path
        self.metrics = metrics

        # prepare output_path dir (if not already exists)
        file.create_dir(os.path.dirname(self.output_path))

        # create file on first attempt to write values
        self.first = True

        pass

    def on_epoch_end(self, mode: str, epoch: int, epoch_results: Dict = None) -> None:
        """
        Adds a line to a CSV with the metrics results of self.metrics.
        The CSV is saved under self.output_path

        :param mode: either 'train', 'validation' or 'infer'.
        :param epoch: epoch number
        :param epoch_results: contains the data to be saved to CSV.
            e.g.: {'losses': {'loss1': mean_loss1,
                               'loss2': mean_loss2,
                                'total_loss': mean (loss1 + loss2)}
                    'metrics': {'metric1': epoch_metric1,
                                'metric2': epoch_metric2}}
        """
        if epoch_results is None:
            return

        lgr = logging.getLogger('Fuse')
        flat_results = FuseUtilsHierarchicalDict.flatten(epoch_results)
        # if this is step 0 - pre-training results should be logged in addition to the metrics csv
        if epoch == 0:
            lgr.info(f'Stats for Pre-Training:')
            for evaluator_name, evaluator_value in sorted(flat_results.items()):
                lgr.info(f'{evaluator_name:20} = {str(evaluator_value)}')

        # create a file
        if self.first:
            header_df = pd.DataFrame(columns=['mode', 'epoch'] + sorted(flat_results.keys()))
            header_df.to_csv(self.output_path, header=True, index=False)
            self.first = False

        metrics_df = pd.DataFrame.from_dict(flat_results, orient='index').T
        metrics_df['mode'] = mode
        metrics_df['epoch'] = epoch

        try:
            if self.metrics is None:
                metrics_df[['mode', 'epoch'] + sorted(flat_results.keys())].to_csv(self.output_path, mode='a', header=False, index=False)
            else:
                metrics_df[['mode', 'epoch'] + self.metrics].to_csv(self.output_path, mode='a', header=False, index=False)
        except Exception as e:
            track = traceback.format_exc()
            lgr.error(track)
            lgr.error("Cannot write epoch stats file (maybe it is open in another program?). Will try again next epoch.")

        pass



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import logging
from typing import Optional, Iterable, Any, Tuple

import matplotlib.pyplot as plt

from fuse.data.visualizer.visualizer_base import FuseVisualizerBase
from fuse.utils.utils_hierarchical_dict import FuseUtilsHierarchicalDict
from fuse.utils.utils_logger import log_object_input_state
from fuse.utils.utils_image_processing import FuseUtilsImageProcessing


class FuseVisualizerDefault(FuseVisualizerBase):
    """
    Visualizer for data including single 2D image with optional mask
    """

    def __init__(self, image_name: str, mask_name: Optional[str] = None,
                 label_name: Optional[str] = None, metadata_names: Iterable[str] = tuple(),
                 pred_name: Optional[str] = None):
        """
        :param image_name: hierarchical key name of the image in batch_dict
        :param mask_name:  hierarchical key name of the mask (gt map) in batch_dict.
                           Optional, won't be displayed if not specified.
        :param label_name: hierarchical key name of the to a global label in batch_dict.
                           Optional, won't be displayed if not specified.
        :param metadata_names: list of hierarchical key name of the metadata - will be printed for every sample
        :param pred_name: hierarchical key name of the prediction in batch_dict.
                           Optional, won't be displayed if not specified.
        """
        # log object input state
        log_object_input_state(self, locals())

        # store input parameters
        self.image_pointer = image_name
        self.mask_name = mask_name
        self.label_name = label_name
        self.metadata_pointers = metadata_names
        self.pred_name = pred_name
        self.matching_function = FuseUtilsImageProcessing.match_img_to_input

    def extract_data(self, sample: dict) -> Tuple[Any, Any, Any, Any, Any]:
        """
        extract required data to visualize from sample
        :param sample: global dict of a sample
        :return: image, mask, label, metadata
        """

        # image
        image = FuseUtilsHierarchicalDict.get(sample, self.image_pointer)

        # mask
        if self.mask_name is not None:
            mask = FuseUtilsHierarchicalDict.get(sample, self.mask_name)
        else:
            mask = None

        # label
        if self.label_name is not None:
            label = FuseUtilsHierarchicalDict.get(sample, self.label_name)
        else:
            label = ''

        # mask
        if self.pred_name is not None:
            pred_mask = FuseUtilsHierarchicalDict.get(sample, self.pred_name)
        else:
            pred_mask = None

        # metadata
        metadata = {metadata_ptr: FuseUtilsHierarchicalDict.get(sample, metadata_ptr) for metadata_ptr in
                    self.metadata_pointers}

        return image, mask, label, metadata, pred_mask

    def visualize(self, sample: dict, block: bool = True) -> None:
        """
        visualize sample
        :param sample: batch_dict - to extract the sample from
        :param block: set to False if the process should not be blocked until the plot will be closed
        :return: None
        """
        # extract data
        image, mask, label, metadata, pred_mask = self.extract_data(sample)

        if mask is not None:
            mask = self.matching_function(mask, image)

        if pred_mask is not None:
            pred_mask = self.matching_function(pred_mask, image)

        # visualize
        num_channels = image.shape[0]

        if pred_mask is not None:
            fig, ax = plt.subplots(num_channels, pred_mask.shape[0]+1, squeeze=False)
        else:
            fig, ax = plt.subplots(num_channels, 1, squeeze=False)

        for channel_idx in range(num_channels):
            ax[channel_idx, 0].title.set_text('image (ch %d) (lbl %s)' % (channel_idx, str(label)))

            ax[channel_idx, 0].imshow(image[channel_idx].squeeze(), cmap='gray')
            if mask is not None:
                ax[channel_idx, 0].imshow(mask[channel_idx], alpha=0.3)

            if pred_mask is not None:
                for c_id in range(pred_mask.shape[0]):
                    max_prob = pred_mask[c_id].max()
                    ax[channel_idx, c_id+1].title.set_text('image (ch %d) (max prob %s)' % (channel_idx, str(max_prob)))

                    ax[channel_idx, c_id+1].imshow(image[channel_idx].squeeze(), cmap='gray')
                    ax[channel_idx, c_id+1].imshow(pred_mask[c_id], alpha=0.3)

        lgr = logging.getLogger('Fuse')
        lgr.info('------------------------------------------')
        lgr.info(metadata)
        lgr.info('image label = ' + str(label))
        lgr.info('------------------------------------------')

        mng = plt.get_current_fig_manager()
        mng.resize(*mng.window.maxsize())
        fig.tight_layout()
        plt.show(block=block)

    def visualize_aug(self, orig_sample: dict, aug_sample: dict, block: bool = True) -> None:
        """
        Visualise and compare augmented and non-augmented version of the sample
        :param orig_sample: batch_dict to extract the original sample from
        :param aug_sample: batch_dict to extract the augmented sample from
        :param block: set to False if the process should not be blocked until the plot will be closed
        :return: None
        """
        # extract data
        orig_image, orig_mask, orig_label, orig_metadata = self.extract_data(orig_sample)
        aug_image, aug_mask, aug_label, aug_metadata = self.extract_data(aug_sample)

        # visualize
        num_channels = orig_image.shape[0]

        fig, ax = plt.subplots(num_channels, 2, squeeze=False)
        for channel_idx in range(num_channels):
            # orig
            ax[channel_idx, 0].title.set_text('image (ch %d) (lbl %s)' % (channel_idx, str(orig_label)))
            ax[channel_idx, 0].imshow(orig_image[channel_idx].squeeze(), cmap='gray')
            if (orig_mask is not None) and (None not in orig_mask):
                ax[channel_idx, 0].imshow(orig_mask, alpha=0.3)

            # augmented
            ax[channel_idx, 1].title.set_text('image (ch %d) (lbl %s)' % (channel_idx, str(aug_label)))
            ax[channel_idx, 1].imshow(aug_image[channel_idx].squeeze(), cmap='gray')
            if (aug_mask is not None) and (None not in aug_mask):
                ax[channel_idx, 1].imshow(aug_mask, alpha=0.3)
        lgr = logging.getLogger('Fuse')
        lgr.info('------------------------------------------')
        lgr.info("original")
        lgr.info(orig_metadata)
        lgr.info('image label = ' + str(orig_label))
        lgr.info("augmented")
        lgr.info(aug_metadata)
        lgr.info('image label = ' + str(aug_label))
        lgr.info('------------------------------------------')

        mng = plt.get_current_fig_manager()
        mng.resize(*mng.window.maxsize())
        fig.tight_layout()
        plt.show(block=block)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import ast
from typing import Hashable

import pandas as pd

from fuse.data.processor.processor_base import FuseProcessorBase


class FuseProcessorCSV(FuseProcessorBase):
    """
    Processor reading data from csv file.
    Covert each row to a dictionary
    """

    def __init__(self, csv_filename: str, sample_desc_column='descriptor'):
        """
        Processor reading data from csv file.
        :param csv_filename: path to the csv file
        :param sample_desc_column: name of the sample descriptor column within the csv file
        """
        self.sample_desc_column = sample_desc_column
        self.csv_filename = csv_filename
        # read csv
        self.data = pd.read_csv(csv_filename)

    def __call__(self, sample_desc: Hashable):
        """
        See base class
        """
        # locate the required item
        items = self.data.loc[self.data[self.sample_desc_column] == str(sample_desc)]
        # convert to dictionary - assumes there is only one item with the requested descriptor
        sample_data = items.to_dict('records')[0]
        for key in sample_data.keys():
            if 'output' in key:
                if isinstance(sample_data[key], str):
                    tuple_data = sample_data[key]
                    if tuple_data.startswith('[') and tuple_data.endswith(']'):
                        sample_data[key] = ast.literal_eval(tuple_data.replace(" ", ","))

        return sample_data



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Union, Sequence

import numpy as np
import torch


def calculate_box_iou(box1: Union[np.ndarray, torch.Tensor],
                      box2: Union[np.ndarray, torch.Tensor],
                      api: str = 'np') -> np.ndarray:
    """
    Calculates intersection-over-union for two sets of bounding boxes, for all possible pairs.
    This method supports both NumPy and PyTorch tensor inputs - remember to specify in the 'api' param.
    :param box1: list of bounding boxes, shape [N, 4+] where first four entries in each row are top-left-bottom-right coordinates.
                 rows may contain more than four entries, only the first four are used.
    :param box2: same as box1 (e.g., use box1 for preds and box2 for targets)
    :param api:  select either 'np' or 'torch'
    :return: A NumPy matrix of IOU for all pair combinations of box1/box2. rows=box1, cols=box2
    """

    assert api in ['np', 'torch']

    if api == 'np':
        n = int(box1.shape[0])
        m = int(box2.shape[0])
        lt = np.maximum(
            np.repeat(box1[:, :2][:, None, :], m, axis=1),  # [n, 2] --> [n, m, 2]
            np.repeat(box2[:, :2][None, :, :], n, axis=0)  # [m, 2] --> [n, m, 2]
        )
        rb = np.minimum(
            np.repeat(box1[:, 2:4][:, None, :], m, axis=1),
            np.repeat(box2[:, 2:4][None, :, :], n, axis=0)
        )
        wh = (rb - lt).clip(min=0)  # [N,M,2]

    elif api == 'torch':
        n = box1.size(0)
        m = box2.size(0)
        lt = torch.max(
            box1[:, :2].unsqueeze(1).expand(n, m, 2),  # [n, 2] -> [n, 1, 2] -> [n, m, 2]
            box2[:, :2].unsqueeze(0).expand(n, m, 2),  # [m ,2] -> [1, m, 2] -> [n, m, 2]
        )
        rb = torch.min(
            box1[:, 2:4].unsqueeze(1).expand(n, m, 2),
            box2[:, 2:4].unsqueeze(0).expand(n, m, 2),
        )
        wh = (rb - lt).clamp(min=0)  # [N,M,2]

    inter = wh[:, :, 0] * wh[:, :, 1]  # [N,M]
    area1 = (box1[:, 2] - box1[:, 0]) * (box1[:, 3] - box1[:, 1])  # [N,]
    area2 = (box2[:, 2] - box2[:, 0]) * (box2[:, 3] - box2[:, 1])  # [M,]

    if api == 'np':
        area1 = np.repeat(area1[:, None], m, axis=1)
        area2 = np.repeat(area2[None, :], n, axis=0)
    elif api == 'torch':
        area1 = area1.unsqueeze(1).expand_as(inter)  # [N,] -> [N,1] -> [N,M]
        area2 = area2.unsqueeze(0).expand_as(inter)  # [M,] -> [1,M] -> [N,M]

    iou = inter / (area1 + area2 - inter)

    if api == 'torch':
        iou = iou.cpu().numpy()

    return iou


def box_non_maximum_suppression(boxes: np.ndarray, scores: np.ndarray, overlap_threshold: float = 0.5) -> Sequence:
    """
    Bounding box non-maximum-suppression.
    Suppresses boxes with intersection-over-union above overlap_threshold, taking the top-scoring overlap box according
    to 'scores' param.

    Adapted and modified from:
    https://www.computervisionblog.com/2011/08/blazing-fast-nmsm-from-exemplar-svm.html
    https://www.pyimagesearch.com/2015/02/16/faster-non-maximum-suppression-python/

    :param boxes:               ndarray, shaped [num_boxes, 4], entries in each row are top-left-bottom-right coordinates.
    :param scores:              ndarray, shaped [num_boxes, ]
    :param overlap_threshold:   threshold for IoU suppression

    :return:  list of indices of non-suppressed boxes
    """

    if len(boxes) == 0:
        return []

    # if the bounding boxes integers, convert them to floats --
    # this is important since we'll be doing a bunch of divisions
    if boxes.dtype.kind == "i":
        boxes = boxes.astype("float")

    # initialize the list of picked indexes
    pick = []

    # grab the coordinates of the bounding boxes
    x1 = boxes[:, 0]
    y1 = boxes[:, 1]
    x2 = boxes[:, 2]
    y2 = boxes[:, 3]

    # compute areas
    area = (x2 - x1 + 1) * (y2 - y1 + 1)

    # sort by scores, descending order
    idxs = np.argsort(scores)[::-1]

    # keep looping while some indexes still remain in the indexes list
    while len(idxs) > 0:
        # grab the last index in the indexes list and add the index value to the list of picked indexes
        last = len(idxs) - 1
        i = idxs[last]
        pick.append(i)

        # find the largest (x, y) coordinates for the start of
        # the bounding box and the smallest (x, y) coordinates
        # for the end of the bounding box
        xx1 = np.maximum(x1[i], x1[idxs[:last]])
        yy1 = np.maximum(y1[i], y1[idxs[:last]])
        xx2 = np.minimum(x2[i], x2[idxs[:last]])
        yy2 = np.minimum(y2[i], y2[idxs[:last]])

        # compute the width and height of the bounding box
        w = np.maximum(0, xx2 - xx1 + 1)
        h = np.maximum(0, yy2 - yy1 + 1)

        # compute the ratio of overlap
        overlap = (w * h) / area[idxs[:last]]

        idxs = np.delete(idxs, np.concatenate(([last],
                                               np.where(overlap >= overlap_threshold)[0])))

    return pick


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    import matplotlib.patches as patches

    visualize = True

    preds = [[100, 250, 400, 450, 0, 0.9],
             [300, 700, 700, 900, 0, 0.8],
             [600, 400, 1000, 1000, 0, 0.4]]
    targets = [[150, 250, 500, 500, 0],
               [200, 710, 880, 950, 0],
               [530, 320, 920, 1100, 0]]

    preds = np.array(preds)
    targets = np.array(targets)
    ious = calculate_box_iou(preds, targets, api='np')
    print('np')
    print(ious)
    ious = calculate_box_iou(torch.Tensor(preds), torch.Tensor(targets), api='torch')
    print('torch')
    print(ious)

    # Visualize data
    if visualize:
        fig, ax = plt.subplots(1)
        ax.imshow(np.zeros([1200, 1200]), cmap='gray')
        for idx, bb in enumerate(preds):
            rect = patches.Rectangle((bb[0], bb[1]), bb[2] - bb[0], bb[3] - bb[1], edgecolor='red', facecolor='none')
            ax.add_patch(rect)
            text_val = ', '.join(["%.2f" % score for score in ious[idx] if score > 0])
            ax.text(bb[0] + 10, bb[1] - 10, text_val, color='red')
        for idx, bb in enumerate(targets):
            rect = patches.Rectangle((bb[0], bb[1]), bb[2] - bb[0], bb[3] - bb[1], edgecolor='green', facecolor='none')
            ax.add_patch(rect)

        plt.show()

    # Test non maximum suppression
    all_boxes = np.array([box[:4] for box in preds] + [box[:4] for box in targets]).astype('int')
    all_scores = np.array([0.9, 0.8, 0.4, 1.0, 1.0, 1.0])

    nms_indices = box_non_maximum_suppression(all_boxes, all_scores, overlap_threshold=0.5)
    print('\nAfter non-maximum-suppression')
    print(all_boxes[nms_indices])
    print(all_scores[nms_indices])




"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import logging
import os
import traceback
from typing import Dict, Optional

import pandas as pd
from torch.utils.data import DataLoader
from tqdm import trange

from fuse.data.data_source.data_source_from_list import FuseDataSourceFromList
from fuse.data.dataset.dataset_base import FuseDatasetBase
from fuse.data.dataset.dataset_default import FuseDatasetDefault
from fuse.data.processor.processor_base import FuseProcessorBase
from fuse.data.processor.processor_dataframe import FuseProcessorDataFrame
from fuse.metrics.metric_base import FuseMetricBase
from fuse.utils.utils_debug import FuseUtilsDebug
from fuse.utils.utils_file import FuseUtilsFile
from fuse.utils.utils_hierarchical_dict import FuseUtilsHierarchicalDict


class FuseAnalyzerDefault:
    """
    Default implementation of generic analyzer
    """

    def __init__(self):
        pass

    def analyze(self, gt_processors: Dict[str, FuseProcessorBase], metrics: Dict[str, FuseMetricBase],
                data: Optional[pd.DataFrame] = None, data_pickle_filename: Optional[str] = None,
                output_filename: Optional[str] = None, **kwargs) -> Dict:
        """
        analyze all the samples included in the infer file.
        :param gt_processors: processors that generates the ground truth data required for analyzing
        :param metrics: dictionary of required metrics
        :param data:  input DataFrame
        :param data_pickle_filename: path to a pickled DataFrame (possible gzipped)
        :param output_filename: path used to save the report
        :param kwargs: additional parameters specified in function analyze_generic()
        :return: collected results
        """
        # create inference processor
        infer_processor = FuseProcessorDataFrame(data=data, data_pickle_filename=data_pickle_filename)

        # create data source
        descriptors_list = infer_processor.get_samples_descriptors()
        data_source = FuseDataSourceFromList(descriptors_list)

        # create dataset
        dataset = FuseDatasetDefault(cache_dest=None,
                                     data_source=data_source,
                                     input_processors={},
                                     gt_processors=gt_processors)
        dataset.create()

        # run analyze generic
        return self.analyze_generic(dataset, infer_processor, metrics, output_filename, **kwargs)

    def analyze_generic(self, dataset: FuseDatasetBase, infer_processor: FuseProcessorBase, metrics: Dict[str, FuseMetricBase],
                        output_filename: Optional[str] = None, output_file_mode: str = 'w', print_results: bool = True,
                        num_workers: int = 4, batch_size: int = 2) -> Dict:
        """
        analyze all the samples in the dataset.
        :param dataset: dataset object - does not need to include the 'input' data.
                        For example DataSetDefault instance can be created only with the ground truth processors
        :param infer_processor: processor extracting the data generated by manager.infer()
                                Typically will be instance of FuseProcessorCSV
        :param metrics: dictionary of required metrics
        :param output_filename: path used to save the report
        :param output_file_mode: either write mode ('w') or append model ('a')
        :param num_workers: number of dataloader workers
        :param batch_size: batch_size - data is loaded and collected in batches
        :param print_results: print metrics results to screen
        :return: collected results
        """
        # debug - num workers
        override_num_workers = FuseUtilsDebug().get_setting('manager_override_num_dataloader_workers')
        if override_num_workers != 'default':
            num_workers = override_num_workers
            logging.getLogger('Fuse').info(f'Manager - debug mode - override dataloader num_workers to {override_num_workers}', {'color': 'red'})

        # create dataloader
        dataloader = DataLoader(dataset=dataset,
                                shuffle=False, drop_last=False,
                                batch_size=batch_size,
                                collate_fn=dataset.collate_fn,
                                num_workers=num_workers)

        # reset metrics
        for metric in metrics.values():
            metric.reset()

        # iterate and collect batches
        data_iter = iter(dataloader)
        for _ in trange(len(dataloader)):
            batch_dict = next(data_iter)
            self.analyze_batch(batch_dict, infer_processor, metrics)

        # metric process
        metrics_results = {metric_name: metric.process() for metric_name, metric in metrics.items()}

        # print results
        if print_results:
            self.print_metrics_results(metrics_results)

        lgr = logging.getLogger('Fuse')

        # dump metrics results
        if output_filename is not None:
            self.dump_metrics_results(metrics_results, output_filename, output_file_mode)
            lgr.info(f'\nAnalyzer done. Results saved in {output_filename}  ', {'color': 'magenta', 'attrs': 'bold'})

        lgr.info(f'\nAnalyzer done.', {'color': 'magenta', 'attrs': 'bold'})
        return metrics_results

    def analyze_batch(self, batch_dict: dict, infer_processor: FuseProcessorBase,
                      metrics: Dict[str, FuseMetricBase]) -> None:
        """
        Static function analyzing batch of samples
        For details about the input parameters can be found in FuseAnalyzerDefault.analyze_generic()
        """
        try:
            # get the sample descriptor of the sample
            sample_descriptors = FuseUtilsHierarchicalDict.get(batch_dict, 'data.descriptor')
            # get the infer data
            infer_batch_data = [infer_processor(descr) for descr in sample_descriptors]
            # add infer data to batch_dict
            for key in infer_batch_data[0]:
                FuseUtilsHierarchicalDict.set(batch_dict, key, [infer_sample_data[key] for infer_sample_data in infer_batch_data])

            # metric collect operation
            for metric in metrics.values():
                metric.collect(batch_dict)
        except:
            lgr = logging.getLogger('Fuse')
            # do not stop analyzing - print the error and continue
            track = traceback.format_exc()
            lgr.error(track)
            lgr.error(f'Failed to analyze batch, skipping')

    def print_metrics_results(self, metrics_results: Dict) -> None:
        """
        Print the metrics results
        :param metrics_results: results return from metric.process()
        :return: None
        """
        lgr = logging.getLogger('Fuse')
        lgr.info(f'Results', {'attrs': ['bold', 'underline']})
        for metric_name, metric_result in metrics_results.items():
            lgr.info(f'\nMetric {metric_name}:', {'attrs': 'bold'})
            if isinstance(metric_result, dict):
                keys = FuseUtilsHierarchicalDict.get_all_keys(metric_result)
                for key in keys:
                    lgr.info(f'{key}: {FuseUtilsHierarchicalDict.get(metric_result, key)}')
            else:
                lgr.info(metric_result)

    def dump_metrics_results(self, metrics_results: Dict, filename: str, output_file_mode: str = 'w') -> None:
        """
        Dump results to a file
        :param metrics_results: results return from metric.process()
        :param filename: output file name
        :param output_file_mode: either write mode ('w') or append model ('a')
        :return: None
        """
        # make sure the output folder exist
        dir_path = os.path.dirname(filename)
        FuseUtilsFile.create_dir(dir_path)

        with open(filename, output_file_mode) as output_file:
            output_file.write(f'Results:\n')

            for metric_name, metric_result in metrics_results.items():
                output_file.write(f'\nMetric {metric_name}:\n')
                output_file.write(f'------------------------------------------------\n')
                if isinstance(metric_result, dict):
                    keys = FuseUtilsHierarchicalDict.get_all_keys(metric_result)
                    for key in keys:
                        output_file.write(f'{key}: {FuseUtilsHierarchicalDict.get(metric_result, key)}\n')
                else:
                    output_file.write(f'{metric_result}\n')



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Optional

import pandas as pd
import numpy as np
import pickle

from fuse.data.data_source.data_source_base import FuseDataSourceBase


class FuseSkinDataSource(FuseDataSourceBase):
    def __init__(self,
                 input_source: str,
                 partition_file: Optional[str] = None,
                 train: bool = True,
                 portion_train: float = 0.7,
                 override_partition: bool = True
                 ):

        """
        Create DataSource
        :param input_source:       path to images
        :param partition_file:     Optional, name of a pickle file when no validation set is available
                                   If train = True, train/val indices are dumped into the file,
                                   If train = False, train/val indices are loaded
        :param train:              specifies if we are in training phase
        :param portion_train:      train proportion in case of splitting
        :param override_partition: specifies if the given partition file is filled with new train/val splits
        """

        # Load csv file
        # ----------------------
        input_df = pd.read_csv(input_source)

        # Extract entities
        # ----------------

        sample_descs = []

        if partition_file is not None:
            if train:
                if override_partition:
                    num_sequences = input_df.shape[0]
                    break_train = int(num_sequences * portion_train)
                    splits = np.random.permutation(np.arange(num_sequences))
                    splits = np.split(splits, [break_train])
                    splits = {'train': splits[0], 'val': splits[1]}

                    with open(partition_file, "wb") as pickle_out:
                        pickle.dump(splits, pickle_out)

                    for sample_id in input_df.iloc[splits['train'], 0]:
                        sample_descs.append(sample_id)
                else:
                    # read from a previous train/test split to evaluate on the same partition
                    with open(partition_file, "rb") as splits:
                        repartition = pickle.load(splits)
                        for sample_id in input_df.iloc[repartition['train'], 0]:
                            sample_descs.append(sample_id)

            else:
                with open(partition_file, "rb") as splits:
                    repartition = pickle.load(splits)
                    for sample_id in input_df.iloc[repartition['val'], 0]:
                        sample_descs.append(sample_id)

        else:
            for sample_id in input_df.iloc[:, 0]:
                sample_descs.append(sample_id)

        self.samples = sample_descs

        self.input_source = input_source

    def get_samples_description(self):
        """
        Returns a list of samples ids.
        :return: list[str]
        """
        return self.samples

    def summary(self) -> str:
        """
        Returns a data summary.
        :return: str
        """
        summary_str = ''
        summary_str += 'Class = FuseSkinDataSource\n'

        if isinstance(self.input_source, str):
            summary_str += 'Input source filename = %s\n' % self.input_source

        summary_str += 'Number of samples = %d\n' % len(self.samples)

        return summary_str



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import unittest
import os
import logging
from fuse.utils.utils_logger import fuse_logger_start

from fuse.managers.manager_default import FuseManagerDefault
from fuse.utils.utils_file import FuseUtilsFile


class FuseManagerTestCase(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):

        FuseUtilsFile.create_or_reset_dir('/tmp/test_manager', force_reset=True)
        fuse_logger_start(output_path='/tmp/test_manager', console_verbose_level=logging.INFO)

        self.manager = FuseManagerDefault('/tmp/test_manager', force_reset=True)
        self.train_dict = {'metric_1': 100, 'metric_2': 80, 'metric_3': 75}
        self.validation_dict = {'metric_1': 90, 'metric_2': 70, 'metric_3': 60}
        self.manager.current_epoch = 7
        pass

    def read_file(self):
        with open("/tmp/test_manager/last_epoch_summary.txt", 'r') as sfile:
            summ = sfile.read()
        return summ

    def test_epoch_summary_one_source(self):
        self.manager.best_epoch_function = ['metric_1']
        self.manager.best_epoch = [5]
        self.manager.best_epoch_values = [{'metric_1': 85, 'metric_2': 84, 'metric_3': 71}]
        self.manager._write_epoch_summary_table(self.train_dict, self.validation_dict, 0)

        summary = self.read_file()
        self.assertTrue('| metric_1                 | 85.0000                  | 90.0000                  | 100.0000                 |' in summary)
        self.assertTrue('| metric_2                 | 84.0000                  | 70.0000                  | 80.0000                  |' in summary)
        self.assertTrue('| metric_3                 | 71.0000                  | 60.0000                  | 75.0000                  |' in summary)
        self.assertTrue('Stats for epoch: 7 (Best epoch is 5 for source metric_1)' in summary)

    def test_epoch_summary_two_sources(self):
        self.manager.best_epoch_function = ['metric_1', 'metric_2']
        self.manager.best_epoch = [5, 12]
        self.manager.best_epoch_values = [{'metric_1': 85, 'metric_2': 84, 'metric_3': 71}, {'metric_1': 81, 'metric_2': 89, 'metric_3': 65}]
        self.manager._write_epoch_summary_table(self.train_dict, self.validation_dict, 0)

        summary = self.read_file()
        self.assertTrue('| metric_1                 | 85.0000                  | 90.0000                  | 100.0000                 |' in summary)
        self.assertTrue('| metric_2                 | 84.0000                  | 70.0000                  | 80.0000                  |' in summary)
        self.assertTrue('| metric_3                 | 71.0000                  | 60.0000                  | 75.0000                  |' in summary)
        self.assertFalse('| metric_3                 | 65.0000                  | 60.0000                  | 75.0000                  |' in summary)
        self.assertTrue('Stats for epoch: 7 (Best epoch is 5 for source metric_1)' in summary)

        self.manager._write_epoch_summary_table(self.train_dict, self.validation_dict, 1)
        summary = self.read_file()
        # see that index 0 source was not deleted
        self.assertTrue('| metric_1                 | 85.0000                  | 90.0000                  | 100.0000                 |' in summary)
        self.assertTrue('| metric_2                 | 84.0000                  | 70.0000                  | 80.0000                  |' in summary)
        self.assertTrue('| metric_3                 | 71.0000                  | 60.0000                  | 75.0000                  |' in summary)
        # see that index 1 was added
        self.assertTrue('| metric_1                 | 81.0000                  | 90.0000                  | 100.0000                 |' in summary)
        self.assertTrue('| metric_2                 | 89.0000                  | 70.0000                  | 80.0000                  |' in summary)
        self.assertTrue('| metric_3                 | 65.0000                  | 60.0000                  | 75.0000                  |' in summary)
        self.assertTrue('Stats for epoch: 7 (Best epoch is 12 for source metric_2)' in summary)

    def test_epoch_summary_none_values(self):
        self.manager.best_epoch_function = ['metric_1']
        self.manager.best_epoch = [5]
        train_dict = {'metric_1': 100, 'metric_2': 80, 'metric_3': None}
        validation_dict = {'metric_1': 90, 'metric_2': None, 'metric_3': None}
        self.manager.best_epoch_values = [{'metric_1': None, 'metric_2': 84, 'metric_3': None}]
        self.manager._write_epoch_summary_table(train_dict, validation_dict, 0)

        summary = self.read_file()
        self.assertTrue('| metric_1                 | N/A                      | 90.0000                  | 100.0000                 |' in summary)
        self.assertTrue('| metric_2                 | 84.0000                  | N/A                      | 80.0000                  |' in summary)
        self.assertTrue('| metric_3                 | N/A                      | N/A                      | N/A                      |' in summary)
        self.assertTrue('Stats for epoch: 7 (Best epoch is 5 for source metric_1)' in summary)

    def test_epoch_summary_string_values(self):
        self.manager.best_epoch_function = ['metric_1']
        self.manager.best_epoch = [5]
        train_dict = {'metric_1': 100, 'metric_2': 80, 'metric_3': 'lala'}
        validation_dict = {'metric_1': 90, 'metric_2': 'lili', 'metric_3': 'lolo'}
        self.manager.best_epoch_values = [{'metric_1': 14, 'metric_2': 84, 'metric_3': 'kiki'}]
        self.manager._write_epoch_summary_table(train_dict, validation_dict, 0)

        summary = self.read_file()
        self.assertTrue('| metric_1                 | 14.0000                  | 90.0000                  | 100.0000                 |' in summary)
        self.assertTrue('| metric_2                 | 84.0000                  | N/A                      | 80.0000                  |' in summary)
        self.assertTrue('| metric_3                 | N/A                      | N/A                      | N/A                      |' in summary)
        self.assertTrue('Stats for epoch: 7 (Best epoch is 5 for source metric_1)' in summary)

    def test_is_best_epoch_so_far(self):
        print("test_is_best_epoch_so_far")

        self.manager.on_equal_values = ['better', 'worse']
        self.manager.optimization_function = ['max', 'max']
        self.manager.best_epoch_function = ['metric_1', 'metric_2']
        self.manager.best_epoch_values = [{'metric_1': 85, 'metric_2': 84, 'metric_3': 71},
                                          {'metric_1': 81, 'metric_2': 89, 'metric_3': 65}]
        train_dict = {}
        validation_dict = {'metric_1': 90, 'metric_2': None, 'metric_3': 'lolo'}

        is_better = self.manager._is_best_epoch_so_far(train_dict, validation_dict, 0)
        self.assertTrue(is_better)
        self.assertDictEqual(self.manager.best_epoch_values[0], validation_dict)

        is_better = self.manager._is_best_epoch_so_far(train_dict, validation_dict, 1)
        self.assertFalse(is_better)
        self.assertDictEqual(self.manager.best_epoch_values[0], {'metric_1': 90, 'metric_2': None, 'metric_3': 'lolo'})

        validation_dict = {'metric_1': 90, 'metric_2': 89, 'metric_3': 'lolo'}
        is_better = self.manager._is_best_epoch_so_far(train_dict, validation_dict, 1)
        self.assertFalse(is_better)
        self.assertDictEqual(self.manager.best_epoch_values[1], {'metric_1': 81, 'metric_2': 89, 'metric_3': 65})

        validation_dict = {'metric_1': 90, 'metric_2': 89.0001, 'metric_3': 'lolo'}
        is_better = self.manager._is_best_epoch_so_far(train_dict, validation_dict, 1)
        self.assertTrue(is_better)
        self.assertDictEqual(self.manager.best_epoch_values[1], validation_dict)

    def test_is_best_epoch_so_far_invalid_metric(self):
        print("test_is_best_epoch_so_far")

        self.manager.on_equal_values = ['better', 'worse']
        self.manager.optimization_function = ['max', 'max']
        self.manager.best_epoch_function = ['wrong_metric', 'metric_2']
        self.manager.best_epoch_values = [{'metric_1': 85, 'metric_2': 84, 'metric_3': 71},
                                          {'metric_1': 81, 'metric_2': 89, 'metric_3': 65}]
        train_dict = {}
        validation_dict = {'metric_1': 90, 'metric_2': 95, 'metric_3': 'lolo'}

        with self.assertRaises(KeyError):
            self.manager._is_best_epoch_so_far(train_dict, validation_dict, 0)

        is_better = self.manager._is_best_epoch_so_far(train_dict, validation_dict, 1)
        self.assertTrue(is_better)
        self.assertDictEqual(self.manager.best_epoch_values[1], validation_dict)

    def tearDown(self):
        pass


if __name__ == '__main__':
    unittest.main()



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import torch.nn as nn


class ClassifierFCN(nn.Module):
    def __init__(self, in_ch, num_classes, layers_description=(256,), dropout_rate=0.1):
        super(ClassifierFCN, self).__init__()
        layer_list = []
        layer_list.append(nn.Conv2d(in_ch, layers_description[0], kernel_size=1, stride=1))
        layer_list.append(nn.ReLU())
        if dropout_rate is not None and dropout_rate > 0:
            layer_list.append(nn.Dropout(p=dropout_rate))
        last_layer_size = layers_description[0]
        for curr_layer_size in layers_description[1:]:
            layer_list.append(nn.Conv2d(last_layer_size, curr_layer_size, kernel_size=1, stride=1))
            layer_list.append(nn.ReLU())
            if dropout_rate is not None and dropout_rate > 0:
                layer_list.append(nn.Dropout(p=dropout_rate))
            last_layer_size = curr_layer_size

        layer_list.append(nn.Conv2d(last_layer_size, num_classes, kernel_size=1, stride=1))
        self.classifier = nn.Sequential(*layer_list)

    def forward(self, x):
        x = self.classifier(x)
        return x



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import logging

import pandas as pd
from typing import Sequence, Hashable, Union, Optional, List, Dict

from fuse.data.data_source.data_source_base import FuseDataSourceBase
from fuse.utils.utils_misc import autodetect_input_source


class FuseDataSourceDefault(FuseDataSourceBase):
    """
    DataSource for the following aut-detectable types:

    1. DataFrame (instance or path to pickled object)
    2. Python list of sample descriptors
    3. Text file (needs to end with '.txt' or '.text' extension)

    """

    def __init__(self, input_source: Union[str, pd.DataFrame, Sequence[Hashable]] = None,
                 folds: Optional[Union[int, Sequence[int]]] = None, conditions: Optional[List[Dict[str, List]]] = None) -> None:
        """
        :param input_source: auto-detectable input source
        :param folds:   if input is a DataFrame having a 'fold' column, filter by this fold(s)
        :param conditions: conditions to apply on data source.
                the conditions are column names that are expected to be in input_source data frame.

                Structure:
                   * List of 'Filter Queries' with logical OR between them.
                   * Each Filter Query is a dictionary of data source column and a list of possible values, with logical AND between the keys.

                Example - selecting only negative or positive biopsy samples:
                   [{'biopsy' : ['positive', 'negative']}]
                Example - selecting negative or positive biopsy biopsy samples that are of type 'tumor':
                   [{'biopsy': ['positive', 'negative'], 'type': ['tumor']}]
                Example - selecting negative/positive biopsy samples that are of type 'calcification' AND marked as BIRAD 0 or 5:
                   [{'biopsy': ['positive', 'negative'], 'type': ['calcification'], 'birad': ['BIRAD0', 'BIRAD5']}]
                Example - selecting samples that are either positive biopsy OR marked as BIRAD 0:
                   [{'biopsy': ['positive']}, {'birad': ['BIRAD0']}]

        """
        self.samples_df = autodetect_input_source(input_source)

        if conditions is not None:
            before = len(self.samples_df)
            to_keep = self.filter_by_conditions(self.samples_df, conditions)
            self.samples_df = self.samples_df[to_keep].copy()
            logging.getLogger('Fuse').info(f"Remove {before - len(self.samples_df)} records that did not meet conditions")

        if self.samples_df is None:
            raise Exception('Error detecting input source in FuseDataSourceDefault')

        if isinstance(folds, int):
            self.folds = [folds]
        else:
            self.folds = folds

        if self.folds is not None:
            assert 'fold' in self.samples_df, f'Data cannot be filtered by folds {folds} as folds are specified in the collected data'
            self.samples_df = self.samples_df[self.samples_df['fold'].isin(self.folds)]

    @staticmethod
    def filter_by_conditions(samples: pd.DataFrame, conditions: Optional[List[Dict[str, List]]]):
        """
        Returns a vector of the samples that passed the conditions
        :param samples: dataframe to check. expected to have at least sample_desc column.
        :param conditions: list of dictionaries. each dictionary has column name as keys and possible values as the values.
                for each dict in the list:
                    the keys are applied with AND between them.
                the dict conditions are applied with OR between them.
        :return: boolean vector with the filtered samples
        """
        to_keep = samples.sample_desc.isna()  # start with all false
        for condition_list in conditions:
            condition_to_keep = samples.sample_desc.notna()  # start with all true
            for column, values in condition_list.items():
                condition_to_keep = condition_to_keep & samples[column].isin(values)  # all conditions in list must be met
            to_keep = to_keep | condition_to_keep  # add this condition samples to_keep
        return to_keep

    def get_samples_description(self):
        return list(self.samples_df['sample_desc'])

    def summary(self) -> str:
        summary_str = ''
        summary_str += 'FuseDataSourceDefault - %d samples\n' % len(self.samples_df)
        return summary_str


if __name__ == '__main__':
    my_df = pd.DataFrame({'sample_desc': range(11, 16),
                          'A': range(1, 6),
                          'B': range(10, 0, -2),
                          'C': range(10, 5, -1)})
    print(my_df)
    clist = [{'A': [2, 3, 4], 'B': [8, 2]}, {'C': [8, 7]}]
    to_keep = FuseDataSourceDefault.filter_by_conditions(my_df, clist)
    print(my_df[to_keep])

    to_keep = FuseDataSourceDefault.filter_by_conditions(my_df, [{}])
    print(my_df[to_keep])



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import pickle
import pandas as pd
import os

class FuseProstateXUtilsData:
    @staticmethod
    def get_dataset(path_to_db: str,set_type: str, db_ver: int,db_name: str,fold_no: int):
        db_name = os.path.join(path_to_db,f'dataset_{db_name}_folds_ver{db_ver}_seed1.pickle')
        with open(db_name, 'rb') as infile:
            db = pickle.load(infile)


        if set_type == 'train':
            other_folds = list(set(range(0,len(db)))-set([fold_no]))
            for i,f in enumerate(other_folds):
                if i==0:
                    data = db['data_fold' + str(f)]
                else:
                    data = pd.concat([data,db['data_fold' + str(f)]],join='inner')

        elif set_type == 'validation':
            data = db['data_fold'+str(fold_no)]
        elif set_type == 'test':
            data = db['test_data']
        else:
            raise Exception(f'Unexpected set type {set_type}')
        return data

    def get_lesions_prostate_x(data: pd.DataFrame):
        outlier_list = []
        lesion_data = data[~data['Patient ID'].isin(outlier_list)]
        return lesion_data


if __name__ == "__main__":
    path_to_db = '/gpfs/haifa/projects/m/msieve_dev3/usr/Tal/my_research/virtual_biopsy/prostate/experiments/V1/'
    # data = FuseCAPVUtilsData.get_dataset(path_to_db=path_to_db,set_type='train', db_ver=18042021,db_name='tcia',fold_no=0)
    # data_lesion = FuseCAPVUtilsData.get_lesions(data)

    data = FuseProstateXUtilsData.get_dataset(path_to_db=path_to_db, set_type='train', db_ver=29042021, db_name='prostate_x',fold_no=0)
    data_lesion = FuseProstateXUtilsData.get_lesions_prostate_x(data)